"""WakeMax Fix"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.postprocess import default_map

def WakeMax(Y_hat, p, map=default_map):
	"""
	Wake Max Addon. Increase the sensitivity
	of waking due to split of active and quiet waking periods,
	thresholded by NR state probability.

	Fix is conducted by determining when the combined probability
	of the AW and QW states for a given timepoint sum to above the probability
	of the NR state for the given timepoint. When this condition is satisified,
	the prediction is overwritten to the waking state with the higher probability.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples, n_classes)
		The raw prediction probabilities.

	p : array-like, shape=(n_samples,)
		The predictions to process. If no Addon
		processing was done prior to this, `p`
		corresponds with Y_hat.

	map : dict
		Mapping the label values to some
		set of integers.

	Returns
	-------
	p : array-like, shape=(n_samples,)
		The post-processed predictions.
	"""
	for i in range(len(p)):
		if p[i] == map['NR'] and np.sum(Y_hat[i,:map['NR']]) > Y_hat[i,map['NR']]:
			p[i] = np.argmax(Y_hat[i,:map['NR']])
	return p



"""Spectral Entropy Processor"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.io import Dataset
from sleepens.utils.data import signal as s

def spectral_entropy(ds, name=None):
	"""
	Process the epoched power spectral density
	and compute spectral entropy.

	Parameters
	----------
	f : ndarray
		List of arrays of frequencies for each epoch.

	Pxx : ndarray
		List of power spectral densities of `data` at
		each epoch.

	Returns
	-------
	entropies : ndarray, shape=(n_epochs,)
		List of entropies at each epoch.
	"""
	f, Pxx = ds.features, ds.data
	entropies = []
	for e in range(min([len(f), len(Pxx)])):
		entropies.append(s.spectral_entropy(f[e], Pxx[e]))
	entropies = np.array(entropies).reshape(-1,1)
	return Dataset(name=name, features=["ENTROPY"], data=entropies)



"""Cross-Validation"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np
from tqdm import trange
from sklearn import metrics
from sklearn.utils import shuffle as sk_shuffle
from copy import deepcopy

from sleepens.utils import create_random_state, check_XY
from sleepens.ml import check_estimator, check_timeseries_estimator

def cross_validate(estimator, X, Y, cv=5, repeat=1, random_state=None,
					shuffle=True, verbose=0):
	"""
	Conduct a k-fold cross validation.

	Parameters
	----------
	estimator : Classifier
		Estimator to use.

	X : array-like
		Data to use for cross-validation.
		If using a Classifier, shape should be (n_samples, n_features).
		If using a TimeSeriesClassifier, shape should be (n_series, n_samples, n_features).

	Y : array-like
		Target labels to use for cross-validation.
		If using a Classifier, shape should be (n_samples,).
		If using a TimeSeriesClassifier, shape should be (n_series, n_samples,).

	cv : int, default=5
		Number of folds to use in k-fold cross-validation

	repeat : int, default=1
		Number of times to repeat cross-validation to average
		the results.

	random_state : None or int or RandomState, default=None
		Initial seed for the RandomState. If `random_state` is None,
		return the RandomState singleton. If `random_state` is an int,
		return a RandomState with the seed set to the int.
		If `random_state` is a RandomState, return that RandomState.

	shuffle : bool, default=True
		Determine if cross-validation should shuffle the data.

	verbose : int, default=0
		Determines the verbosity of cross-validation.
		Higher verbose levels result in more output logged.

	Returns
	-------
	score : float
		The average score across the cross-validation.

	Y_hat : ndarray or list of ndarray
		The cross-validated predictions.
		If using a Classifier, `Y_hat` is an ndarray with a shape of (n_samples, n_classes)
		If using a TimeSeriesClassifier, `Y_hat` is a list of ndarrays with
		a shape of (n_series, n_samples, n_classes)
	"""
	random_state = create_random_state(random_state)
	check_estimator(estimator)
	if len(X) < cv:
		if len(X) == 1 : raise ValueError("X is length 1 and cross validation requires at least 2 samples/files")
		cv = len(X)
		if verbose > 1 : print("Found too small X, reducing cv to", cv)
	if verbose > 0 : print("Cross-validation with", cv, "folds", "and", repeat, "repeats")
	try:
		check_timeseries_estimator(estimator)
	except : return _cross_validate_classic(estimator, X, Y, cv, repeat,
				random_state, shuffle, verbose)
	else : return _cross_validate_timeseries(estimator, X, Y, cv, repeat,
				random_state, shuffle, verbose)

def _cross_validate_classic(estimator, X, Y, cv, repeat, random_state,
					shuffle, verbose):
	"""
	Conduct a k-fold cross validation on a Classifier.

	Parameters
	----------
	estimator : Classifier
		Estimator to use.

	X : array-like, shape=(n_samples, n_features)
		Data to use for cross-validation.

	Y : array-like, shape=(n_samples,)
		Target labels to use for cross-validation.

	cv : int
		Number of folds to use in k-fold cross-validation

	repeat : int
		Number of times to repeat cross-validation to average
		the results.

	random_state : RandomState
		RandomState for shuffling.

	shuffle : bool
		Determine if cross-validation should shuffle the data.

	verbose : int
		Determines the verbosity of cross-validation.
		Higher verbose levels result in more output logged.

	Returns
	-------
	score : float
		The average score across the cross-validation.

	Y_hat : ndarray, shape=(n_samples, n_classes)
		The cross-validated predictions.
	"""
	e_verbose = estimator.verbose
	suppressed_verbose = verbose - 1 if verbose >= 1 else 0
	estimator.set_verbose(suppressed_verbose)
	X, Y = check_XY(X=X, Y=Y)
	if Y.ndim == 1 : Y = Y.reshape(-1,1)
	Y_hat = np.zeros((len(Y), len(np.unique(Y))))
	for r in range(repeat):
		if verbose > 0 : print("Fitting on repeat", r+1)
		x_y = np.concatenate((X, Y, np.arange(len(X)).reshape(-1,1)), axis=1)
		if shuffle : random_state.shuffle(x_y)
		X_s, Y_s, i_s = x_y[:,:X.shape[-1]], x_y[:,X.shape[-1]:-1], x_y[:,-1]
		X_folds, Y_folds, i_folds = (np.array_split(X_s, cv),
									np.array_split(Y_s, cv),
									np.array_split(i_s, cv))
		if verbose == 1 : folds = trange(cv)
		else : folds = range(cv)
		for f in folds:
			if verbose > 1 : print("Fitting on fold", f+1)
			x_ = np.concatenate(X_folds[:-1]) if len(X_folds) > 2 else X_folds[0]
			y_ = np.concatenate(Y_folds[:-1]) if len(Y_folds) > 2 else Y_folds[0]
			e = deepcopy(estimator)
			e.fit(x_, y_)
			if verbose > 1 : print("Predicting for fold", f+1)
			Y_hat_ = e.predict_proba(X_folds[-1])
			if verbose > 2:
				score_ = metrics.accuracy_score(Y_folds[-1], np.argmax(Y_hat_, axis=1))
				print("Score:", score_)
			Y_hat[i_folds[-1].astype(int)] += Y_hat_ / repeat
			X_folds = X_folds[-1:] + X_folds[:-1]
			Y_folds = Y_folds[-1:] + Y_folds[:-1]
			i_folds = i_folds[-1:] + i_folds[:-1]
	estimator.set_verbose(e_verbose)
	p = np.argmax(Y_hat, axis=1)
	return metrics.accuracy_score(Y, p), Y_hat

def _cross_validate_timeseries(estimator, X, Y, cv, repeat, random_state,
					shuffle, verbose):
	"""
	Conduct a k-fold cross validation on a TimeSeriesClassifier.

	Parameters
	----------
	estimator : TimeSeriesClassifier
		Estimator to use.

	X : list of array-like, shape=(n_series, n_samples, n_features)
		Data to use for cross-validation. Series can be of different lengths.

	Y : list of array-like, shape=(n_series, n_samples,)
		Target labels to use for cross-validation.

	cv : int
		Number of folds to use in k-fold cross-validation

	repeat : int
		Number of times to repeat cross-validation to average
		the results.

	random_state : RandomState
		RandomState for shuffling.

	shuffle : bool
		Determine if cross-validation should shuffle the data.

	verbose : int
		Determines the verbosity of cross-validation.
		Higher verbose levels result in more output logged.

	Returns
	-------
	score : float
		The average score across the cross-validation.

	Y_hat : list of ndarray, shape=(n_series, n_samples, n_classes)
		The cross-validated predictions.
	"""
	e_verbose = estimator.verbose
	suppressed_verbose = verbose - 1 if verbose >= 1 else 0
	estimator.set_verbose(suppressed_verbose)
	for x, y in zip(X, Y) : check_XY(X=x, Y=y)
	try : X_, Y_ = np.concatenate(X), np.concatenate(Y)
	except : raise ValueError("Inputs have different number of features")
	for i in range(len(Y)):
		if Y[i].ndim == 1 : Y[i] = Y[i].reshape(-1,1)
	Y_hat = [np.zeros((len(y), len(np.unique(y)))) for y in Y]
	lens = [len(y) for y in Y]
	for r in range(repeat):
		if verbose > 0 : print("Fitting on repeat", r+1)
		X_s, Y_s, i_s = sk_shuffle(X, Y, np.arange(len(X), dtype=int), random_state=random_state)
		X_folds, Y_folds, i_folds = (np.array_split(np.array(X_s, dtype=object), cv),
									np.array_split(np.array(Y_s, dtype=object), cv),
									np.array_split(np.array(i_s, dtype=object), cv))
		if verbose == 1 : folds = trange(cv)
		else : folds = range(cv)
		for f in folds:
			if verbose > 1 : print("Fitting on fold", f+1)
			x_ = np.concatenate(X_folds[:-1]) if len(X_folds) > 2 else X_folds[0]
			y_ = np.concatenate(Y_folds[:-1]) if len(Y_folds) > 2 else Y_folds[0]
			e = deepcopy(estimator)
			e.fit(x_, y_)
			if verbose > 1 : print("Predicting for fold", f+1)
			Y_hat_ = e.predict_proba(X_folds[-1])
			if verbose > 2:
				scores_, lens_ = [], [len(y_f) for y_f in Y_folds[-1]]
				for y_hat, y in zip(Y_hat_, Y_folds[-1]):
					scores_.append(metrics.accuracy_score(y, np.argmax(y_hat, axis=1)))
				score_ = np.average(scores_, weights=lens_)
				print("Score:", score_)
			for i in range(len(Y_hat_)):
				Y_hat[i_folds[-1][i]] += Y_hat_[i] / repeat
			X_folds = X_folds[-1:] + X_folds[:-1]
			Y_folds = Y_folds[-1:] + Y_folds[:-1]
			i_folds = i_folds[-1:] + i_folds[:-1]
	estimator.set_verbose(e_verbose)
	scores = []
	for y_hat, y in zip(Y_hat, Y):
		scores.append(metrics.accuracy_score(y, np.argmax(y_hat, axis=1)))
	score = np.average(scores, weights=lens)
	return score, Y_hat



from ._check import is_float, is_int, check_XY, check_X, check_Y
from ._random import create_random_state
from ._misc import aggregate, determine_threshold
from ._data_mgmt import one_hot, decode, separate_by_label, time_array, calculate_epochs, get_epoch
from ._calculate import calculate_batch, calculate_weight, calculate_bootstrap

__all__ = [
			'is_float',
			'is_int',
			'check_XY',
			'check_X',
			'one_hot',
			'separate_by_label',
			'time_array',
			'calculate_epochs',
			'get_epoch',
			'aggregate',
			'determine_threshold',
			'create_random_state',
			'calculate_batch',
			'calculate_weight',
			'calculate_bootstrap',
]



from ._gb import GradientBoostingClassifier
from ._tsens import TimeSeriesEnsemble
from ._stsens import StackedTimeSeriesEnsemble

__all__ = [
			'GradientBoostingClassifier',
			'TimeSeriesEnsemble',
			'StackedTimeSeriesEnsemble',
		]



"""MinREM Fix"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.postprocess import default_map

def WakeToREM(Y_hat, p, map=default_map):
	"""
	Wake to REM addon. Brute repair of invalid
	transition from waking to REM sleep.

	Fix is conducted by ablating the entire REM episode proceeding an
	invalid AW/QW to R transition. The post-processed prediction is
	the highest probability of the remaining 3 valid states.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples, n_classes)
		The raw prediction probabilities.

	p : array-like, shape=(n_samples,)
		The predictions to process. If no Addon
		processing was done prior to this, `p`
		corresponds with Y_hat.

	map : dict
		Mapping the label values to some
		set of integers.

	Returns
	-------
	p : array-like, shape=(n_samples,)
		The post-processed predictions.
	"""
	for i in range(len(p)-1):
		if p[i] < map['NR'] and p[i+1] == map['R']:
			j = i+1
			while j < len(p) and p[j] == map['R']:
				j += 1
			p[i+1:j] = np.argmax(Y_hat[i+1:j,:map['R']], axis=1)
	return p



"""Transform Processor"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

from sleepens.io import Dataset

def transform(ds, name=None, ops=[]):
	"""
	Transform feature data according
	to parameters. Transformation
	involves normalization, smoothing,
	and/or transformation.

	Parameters
	----------
	data : ndarray, shape=(n_epochs,)
		Feature data.

	ops : list
		Ordered list of transformation operations
		to perform on `data`. Each operation
		is represented by a dictionary of the function
		followed by applicable arguments.

	Returns
	-------
	transformed : ndarray, shape=(n_epochs,)
		Data after transformation.
	"""
	data = ds.data
	for op in ops:
		func, args, kwargs = None, [], {}
		if 'FUNCTION' in op.keys() : func = op['FUNCTION']
		if 'ARGS' in op.keys() : args = op['ARGS']
		if 'KWARGS' in op.keys() : kwargs = op['KWARGS']
		if func is not None : data = func(data, *args, **kwargs)
	return Dataset(name=name, features=ds.features, data=data)



"""Oversampling Functions"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np
from sklearn.mixture import GaussianMixture as GMM

from sleepens.utils import separate_by_label, create_random_state

def balance(data, labels, desired=None, balance='auto',
							seed=None, verbose=0):
	"""
	Balance a dataset through generative oversampling.

	Model a generative algorithm over each of the
	desired labelled data to sample more data from
	the resulting distributions.

	Parameters
	----------
	data : array-like, shape=(n_samples, n_features)
		Data.

	labels : array-like, shape=(n_samples,)
		Labels corresponding to data.

	desired : {None, array-like}, default=None
		List of labels to oversample.
		If None, balances all labels.

	balance : {'auto', int, dict}, default='auto'
		Determines how to balance and oversample.
		Must be one of:
		 - 'auto' : Automatically balance the dataset
		 			by oversampling `desired` labels to
					match the number of samples in the majority
					label. Recommended to set `desired` to None,
					or risk not oversampling a minority label.
		 - int : Oversample `desired` labels so that total
		 			such samples reach this value.
		 - dict : Oversample each label by the value given.

	seed : None or int or RandomState, default=None
		Initial seed for the RandomState. If seed is None,
		return the RandomState singleton. If seed is an int,
		return a RandomState with the seed set to the int.
		If seed is a RandomState, return that RandomState.

	verbose : int, default=0
		Verbosity; higher values result in
		more verbose output.

	Returns
	-------
	data : ndarray
		All oversampled data, shuffled.

	labels : ndarray
		All corresponding labels.
	"""
	if len(data) == 0:
		if verbose : print("No data to oversample")
		return np.array([]), np.array([])
	if verbose > 0 : print("Balancing Dataset")
	data = separate_by_label(data, labels)
	separated = data
	if desired is not None:
		data_labels = list(data.keys())
		for d in desired:
			if desired not in  data_labels:
				raise ValueError("Desired label is not in the data")
		separated = {d: separated[d] for d in desired}
	if balance == 'auto' or isinstance(balance, int):
		if balance == 'auto':
			if verbose > 0 : print("Balancing Dataset. Method set to 'auto'")
			target = np.max([len(data[k]) for k in data.keys()])
		else:
			if verbose > 0 : print("Balancing Dataset up to", str(balance))
			target = balance
		n_samples = {k: target - len(separated[k]) for k in separated.keys()}
		n_samples = {k: 0 if n_samples[k] < 0 else n_samples[k] for k in separated.keys()}
	elif isinstance(balance, (dict)):
		if verbose > 0 : print("Balancing Dataset. Method set to custom")
		n_samples = balance
	else:
		raise ValueError("Balance must be {'auto', int, dict}")
	if verbose > 0 : print("Oversampling")
	return _generative_oversample(separated, n_samples, seed=seed, verbose=verbose)

def scale(data, labels, factor=1, seed=None, verbose=0):
	"""
	Oversample the dataset.

	Model a generative algorithm over each of the
	desired labelled data to sample more data from
	the resulting distributions.

	Parameters
	----------
	data : array-like, shape=(n_samples, n_features)
		Data.

	labels : array-like, shape=(n_samples,)
		Labels corresponding to data.

	factor : float, default=1
		Factor to oversample the dataset by.

	seed : None or int or RandomState, default=None
		Initial seed for the RandomState. If seed is None,
		return the RandomState singleton. If seed is an int,
		return a RandomState with the seed set to the int.
		If seed is a RandomState, return that RandomState.

	verbose : int, default=0
		Verbosity; higher values result in
		more verbose output.

	Returns
	-------
	data : ndarray
		All oversampled data, shuffled.

	labels : ndarray
		All corresponding labels.
	"""
	if factor <= 0 : raise ValueError("Factor must be positive")
	if len(data) == 0:
		if verbose > 0 : print("No data to oversample")
		return np.array([]), np.array([])
	if verbose > 0 : print("Scaling Dataset")
	separated = separate_by_label(data, labels)
	n_samples = {k: int(factor * len(separated[k])) for k in separated.keys()}
	return _generative_oversample(separated, n_samples, seed=seed, verbose=verbose)

def sample(data, labels, sizes, seed=None, verbose=0):
	"""
	Sample from the dataset.

	Model a generative algorithm over each of the
	desired labelled data to sample more data from
	the resulting distributions.

	Parameters
	----------
	data : array-like, shape=(n_samples, n_features)
		Data.

	labels : array-like, shape=(n_samples,)
		Labels corresponding to data.

	sizes : dict
		Number of samples for each class.

	seed : None or int or RandomState, default=None
		Initial seed for the RandomState. If seed is None,
		return the RandomState singleton. If seed is an int,
		return a RandomState with the seed set to the int.
		If seed is a RandomState, return that RandomState.

	verbose : int, default=0
		Verbosity; higher values result in
		more verbose output.

	Returns
	-------
	data : ndarray
		All oversampled data, shuffled.

	labels : ndarray
		All corresponding labels.
	"""
	if len(data) == 0:
		if verbose > 0 : print("No data to oversample")
		return np.array([]), np.array([])
	if verbose > 0 : print("Sampling Dataset")
	separated = separate_by_label(data, labels)
	return _generative_oversample(separated, sizes, seed=seed, verbose=verbose)

def _generative_oversample(data_labels, n_samples, seed=None, verbose=0):
	"""
	Generatively oversample the data.

	Parameters
	----------
	data_labels : dict
		Dictionary of data categorized by label.

	n_samples : dict
		Dictionary of the number of samples to oversample
		each label.

	seed : None or int or RandomState, default=None
		Initial seed for the RandomState. If seed is None,
		return the RandomState singleton. If seed is an int,
		return a RandomState with the seed set to the int.
		If seed is a RandomState, return that RandomState.

	verbose : int, default=0
		Verbosity; higher values result in
		more verbose output.

	Returns
	-------
	data : ndarray
		All oversampled data, shuffled.

	labels : ndarray
		All corresponding labels.
	"""
	oversampled = {}
	for label in data_labels:
		if n_samples[label] == 0 or len(data_labels[label]) < 2 : continue
		if verbose > 0 : print("\tModelling distribution for", str(label))
		model = _fit_cluster(data_labels[label], seed=seed)
		if verbose > 0 : print("\tSampling data for", str(label))
		model.weights_ = (model.weights_ / np.sum(model.weights_)).astype(np.float64)
		oversampled[label] = model.sample(n_samples[label])[0]
	if verbose > 0 : print("Collating and shuffling")
	new_set = []
	for k in oversampled.keys():
		length = len(oversampled[k])
		labels = np.array([k]*length).reshape(length, 1)
		new_set += list(np.concatenate((oversampled[k], labels), axis=1))
	create_random_state(seed=seed).shuffle(new_set)
	new_set = np.array(new_set)
	if new_set.size == 0 : return np.array([]), np.array([]).astype(int)
	return new_set[:,:-1], new_set[:,-1].astype(int)

def _fit_cluster(data, seed=None):
	"""
	Fit a Gaussian Mixture Model to the given data.

	Parameters
	----------
	data : array-like, shape=(n_samples, n_features)
		Data.

	seed : None or int or RandomState, default=None
		Initial seed for the RandomState. If seed is None,
		return the RandomState singleton. If seed is an int,
		return a RandomState with the seed set to the int.
		If seed is a RandomState, return that RandomState.

	Returns
	-------
	model : GaussianMixture
		The best fitted Gaussian Miture Model as determined
		by the mean of the BIC and AIC for the respective model.
	"""
	data = np.array(data)
	models = []
	abic = []
	n_components = min([len(data), 10])
	for i in range(n_components):
		if len(data) < 2 * (i+1) : continue
		m = GMM(n_components=i+1, n_init=5, random_state=seed)
		m.fit(data)
		models.append(m)
		abic.append(np.mean([m.bic(data), m.aic(data)]))
	return models[np.argmin(abic)]



import numpy as np
from tqdm import trange

from sleepens.ml import cross_validate
from sleepens.ml._base_model import TimeSeriesClassifier
from sleepens.ml.models import TimeSeriesEnsemble

class StackedTimeSeriesEnsemble(TimeSeriesClassifier):
	"""
	A Stacked Time Series Ensemble is an Ensemble Classifier that
	stacks two Time Series Classifiers to better classify time series data.

	Parameters
	----------
	layer_1 : TimeSeriesClassifier, default=TimeSeriesEnsemble
		The first layer of TimeSeriesClassifier.

	layer_2 : TimeSeriesClassifier, default=TimeSeriesEnsemble
		The second layer of TimeSeriesClassifier.

	warm_start : bool, default=False
		Determines warm starting to allow training to pick
		up from previous training sessions.

	random_state : None or int or RandomState, default=None
		Initial seed for the RandomState. If `random_state` is None,
		return the RandomState singleton. If `random_state` is an int,
		return a RandomState with the seed set to the int.
		If seed is a RandomState, return that RandomState.

	verbose : int, default=0
		Verbosity of estimator; higher values result in
		more verbose output.

	Attributes
	----------
	n_classes_ : int
		Number of classes.

	n_features_ : int
		Number of features.
	"""
	def __init__(self, layer_1=TimeSeriesEnsemble(), layer_2=TimeSeriesEnsemble(),
					warm_start=False, random_state=None, verbose=0):
		TimeSeriesClassifier.__init__(self, warm_start=warm_start,
							random_state=random_state, verbose=verbose)
		if not issubclass(type(layer_1), TimeSeriesClassifier):
			raise ValueError("Layer 1 must be a TimeSeriesClassifier")
		if not issubclass(type(layer_2), TimeSeriesClassifier):
			raise ValueError("Layer 2 must be a TimeSeriesClassifier")
		self.layer_1 = layer_1
		self.layer_2 = layer_2
		self.set_warm_start(warm_start)
		self.set_verbose(verbose)
		self.set_random_state(random_state)

	def _initialize(self):
		"""
		Initialize the parameters of the classifier.
		"""
		self.set_warm_start(self.warm_start)

	def fit(self, X, Y):
		"""
		Train the classifier on the given data and labels.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Training data.

		Y : array-like, shape=(n_samples,)
			Target labels as integers.

		Returns
		-------
		self : Classifier
			Fitted classifier.
		"""
		X, Y = self._fit_setup(X, Y)
		if self.verbose > 1 : print("Fitting First Layer")
		score, X_2 = cross_validate(self.layer_1, X, Y, verbose=self.verbose-1)
		self.layer_1.fit(X, Y)
		if self.verbose > 1 : print("Fitting Second Layer")
		self.layer_2.fit(X_2, Y)
		if self.verbose > 1 : print("Completed training")
		return self

	def predict_proba(self, X):
		"""
		Return the prediction probabilities on the given data.

		Parameters
		----------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Data to predict.

		Returns
		-------
		Y_hat : list of ndarray, shape=(s_series, n_samples, n_classes)
			Probability predictions.
		"""
		X = self._predict_setup(X)
		if self.verbose > 1 : print("Predicting on First Layer")
		X_2 = self.layer_1.predict_proba(X)
		if self.verbose > 1 : print("Predicting on Second Layer")
		Y_hat = self.layer_2.predict_proba(X_2)
		if self.verbose > 1 : print("Completed predicting")
		return Y_hat

	def set_verbose(self, verbose):
		"""
		Set the verbosity of the Classifier.
		Subsidiary Classifiers have their verbosity
		suppressed by one compared to this Classifier.

		Parameters
		----------
		verbose : int, default=0
			Determines the verbosity of cross-validation.
			Higher verbose levels result in more output logged.
		"""
		TimeSeriesClassifier.set_verbose(self, verbose)
		for e in (self.layer_1, self.layer_2):
			e.set_verbose(verbose-1)

	def set_random_state(self, random_state):
		"""
		Set the RandomState of the Classifier.
		Subsidiary Classifiers set their RandomState
		based on this RandomState with a seed selected
		from 0 to 2**16.

		Parameters
		----------
		random_state : None or int or RandomState, default=None
			Initial seed for the RandomState. If `random_state` is None,
			return the RandomState singleton. If `random_state` is an int,
			return a RandomState with the seed set to the int.
			If `random_state` is a RandomState, return that RandomState.
		"""
		TimeSeriesClassifier.set_random_state(self, random_state)
		for e in (self.layer_1, self.layer_2):
			e.set_random_state(self.random_state.randint(0, 2**16))

	def set_warm_start(self, warm_start):
		"""
		Set the status of `warm_start` of the Classifier.

		Parameters
		----------
		warm_start : bool
			Determines warm starting to allow training to pick
			up from previous training sessions.
		"""
		TimeSeriesClassifier.set_warm_start(self, warm_start)
		for e in (self.layer_1, self.layer_2):
			e.set_warm_start(warm_start)

	def _is_fitted(self):
		"""
		Returns if the Classifier has been trained and is
		ready to predict new data.

		Returns
		-------
		fitted : bool
			True if the Classifier is fitted, False otherwise.
		"""
		attributes = ["n_classes_","n_features_"]
		return TimeSeriesClassifier._is_fitted(self, attributes=attributes) and \
				self.layer_1._is_fitted() and self.layer_2._is_fitted()



"""Transition Fix"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np
from copy import deepcopy

default = np.array([	[1, 1, 1, 0],
						[1, 1, 1, 0],
						[1, 1, 1, 1],
						[1, 1, 0, 2]])

def TransitionFix(Y_hat, p, transitions=default):
	"""
	Transition Fix Addon. Ensure predictions follow
	physiologically feasible state transitions.

	Fix is conducted when an invalid transition is discovered.
	Half a forward pass is conducted from the invalid transition
	point, adjusting state predictions until the transition error resolves.
	In the forward pass, predictions at time `t` are considered immutable while
	`t+1` can be fixed.

	Subsequently, a backwards pass from the end of the forward pass is done,
	adjusting state predictions until the transition error resolves.
	In the backwards pass, predictions at `t+1` are considered immutable while
	`t` can be fixed.

	Finally, the remaining half of the forward pass is conducted until reaching
	the point of the original error. The pass with the highest internal probability,
	as determined by the transition matrix, is used to determine the new
	state predictions.

	It is recommended this be used as the final Addon to guarantee
	state transitions.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples, n_classes)
		The raw prediction probabilities.

	p : array-like, shape=(n_samples,)
		The predictions to process. If no Addon
		processing was done prior to this, `p`
		corresponds with Y_hat.

	transitions : ndarray, shape=(n_classes, n_classes)
		The matrix of possible transitions. Rows represent
		timepoint `t` while columns represent `t+1`.
		0 represents an invalid transition, while non-zero numbers
		act as a multiplicative factor of the probability to
		determine the most likely transition fix sequence.
		Defaults to:
		[[1, 1, 1, 0],
		 [1, 1, 1, 0],
		 [1, 1, 1, 1],
		 [1, 1, 0, 2]]

	Returns
	-------
	p : array-like, shape=(n_samples,)
		The post-processed predictions.
	"""
	transitions = np.array(transitions)
	for i in range(len(p)-1):
		if transitions[p[i], p[i+1]] == 0:
			k = i
			p_forward, p_f = 0, deepcopy(p)
			while k < len(p)-1 and transitions[p_f[k], p_f[k+1]] == 0:
				p_forward += np.max(Y_hat[k+1]*transitions[p_f[k]])
				p_f[k+1] = np.argmax(Y_hat[k+1]*transitions[p_f[k]])
				k += 1
			k += 1
			end = k
			inverse = transitions.T
			p_back, p_b = 0, deepcopy(p)
			if k == len(p):
				p_back += np.max(Y_hat[k-1])
				p_b[k-1] = np.argmax(Y_hat[k-1])
				k -= 1
			while k > 0 and (k >= i or inverse[p_b[k], p_b[k-1]] == 0):
				p_back += np.max(Y_hat[k-1]*inverse[p_b[k]])
				p_b[k-1] = np.argmax(Y_hat[k-1]*inverse[p_b[k]])
				k -= 1
			start = k
			k -= 1
			while k < i:
				p_forward += np.max(Y_hat[k+1]*transitions[p_f[k]])
				p_f[k+1] = np.argmax(Y_hat[k+1]*transitions[p_f[k]])
				k += 1
			if p_forward >= p_back : p = p_f
			else : p = p_b
	return p



import tkinter as tk
from tkinter import filedialog
import os, joblib, importlib, time, sys, struct
import numpy as np
from tqdm import trange
from pathlib import Path
from pprint import pprint

from sleepens.io import Dataset, DataObject
from sleepens.ml import cross_validate
from sleepens.analysis import confusion_matrix, classification_report
from sleepens.io.interfaces import interfaces

from sleepens.protocols import protocols
from sleepens.protocols.sleepens4 import SleepEnsemble4

version = "1.0.1"
model = SleepEnsemble4()
verbose = 4

def displayMenu():
	print("-"*30)
	print("Sleep Ensemble", version)
	print("Current model:", model.name)
	print("-"*30)
	print("1. Classify some files")
	print("2. Train a new Sleep Ensemble")
	print("3. Validate model")
	print("4. Load model")
	print("5. Export current model")
	print("6. Change verbosity of messages (currently", str(verbose) + ")")
	print("7. Update parameters")
	print("Q. Exit (Emergency abort is Ctrl+C)")
	print("-"*30)

def yes_no_loop(msg):
	while True:
		usr = input(msg + " [y/n]: ").upper()
		if usr == "Y" : return True
		elif usr== "N" : return False
		else : print("Sorry, didn't quite catch that.")

def import_module(name, path):
	try:
		spec = importlib.util.spec_from_file(name, path)
		mod = util.module_from_spec(spec)
		spec.loader.exec_module(mod)
		return mod
	except:
		return importlib.machinery.SourceFileLoader(name, path).load_module()

def ask_filename(title="Choose File", filetypes=[("Any File", "*")]):
	root = tk.Tk()
	root.withdraw()
	filepath = filedialog.askopenfilename(title=title,
					filetypes=filetypes)
	root.destroy()
	return filepath

def ask_filenames(title="Choose Files", filetypes=[("Any File", "*")]):
	root = tk.Tk()
	root.withdraw()
	filepaths = root.tk.splitlist(filedialog.askopenfilenames(title=title,
					filetypes=filetypes))
	root.destroy()
	return filepaths

def save_file(title="Save File", defaultextension=".txt", filetypes=[("Any File", "*")]):
	root = tk.Tk()
	root.withdraw()
	file = filedialog.asksaveasfilename(title=title, filetypes=filetypes)
	if not file : return file
	version = "-py"
	for i in range(3):
		version += str(sys.version_info[i]) + "."
	version += str(sys.hexversion) + "-"
	file += version + sys.platform + "-" + str(struct.calcsize("P")*8) + "bit"
	file += defaultextension
	root.destroy()
	return file

def ask_directory(title="Choose Folder"):
	root = tk.Tk()
	root.withdraw()
	folder = filedialog.askdirectory(title=title)
	root.destroy()
	return folder

def print_report(Y_hat, Y):
	print("Confusion Matrix")
	print("-"*30)
	matrix, targets, labels = confusion_matrix(Y_hat, Y)
	row_format = "{:>10}" * (len(labels) + 2)
	title = ["" for i in range(len(labels)+2)]
	title[len(labels)//2 + 2] = "PREDICTION"
	print(row_format.format(*title))
	print(row_format.format("", "", *labels))
	for row in range(len(matrix)):
		if row == len(matrix)//2 : blank = "TARGET"
		else : blank = ""
		row_data = ["%.f" % cell for cell in matrix[row]]
		print(row_format.format(blank, "%.f" % targets[row], *row_data))
	print("-"*30)
	print("Classification Report")
	print("-"*30)
	report = classification_report(Y_hat, Y)
	beta = report.pop('beta')
	support = report.pop('support')
	accuracy = report.pop('accuracy')
	macro = report.pop('macro avg')
	weighted = report.pop('weighted avg')
	header = ["Precision", "Recall", "F1-Score", "Support"]
	row_format = "{:>15}" * (len(header) + 1)
	print(row_format.format("", *header))
	for k, v in sorted(report.items()):
		print(row_format.format("%.f" % k, "%.4f" % v['precision'], "%.4f" % v['recall'],
								"%.4f" % v['f-score'], "%.f" % v['support']))
	print("")
	print(row_format.format('accuracy', "", "", "%.4f" % accuracy, "%.f" % support))
	print(row_format.format('macro avg', "%.4f" % macro['precision'], "%.4f" % macro['recall'],
							"%.4f" % macro['f-score'], "%.f" % support))
	print(row_format.format('weighted avg', "%.4f" % weighted['precision'], "%.4f" % weighted['recall'],
							"%.4f" % weighted['f-score'], "%.f" % support))
	print("-"*30)

def change_protocol():
	global model, verbose
	while True:
		print("-"*30)
		print("Select Sleep Ensemble Protocol")
		print("-"*30)
		for i in range(len(protocols)):
			print(str(i+1)+".", protocols[i].__name__)
		print("Q. Return to main menu")
		print("-"*30)
		usr = input("Select protocol: ").upper()
		if usr.isdigit():
			usr = int(usr)
			if usr > 0 and usr <= len(protocols):
				protocol = protocols[usr-1]
				print("Changed protocol to", protocol.__name__)
				return protocol
			else : print("Sorry, didn't quite catch that.")
		elif usr == "Q": return
		else : print("Sorry, didn't quite catch that.")

def read_data(in_int, filepath, labels=False):
	global model, verbose
	if verbose > 1 : print("Reading file", filepath)
	if in_int.type == "RAW":
		data, labels = model.read(in_int, filepath, labels=labels)
		name = Path(filepath).stem
		ds = model.process(data, labels, name)
	elif in_int.type == "DATASET":
		ds = in_int.read_data(filepath, model.params['reader']['DATA_COLS'])
		if labels:
			label_ds = in_int.read_labels(filepath, model.params['reader']['LABEL_COLS'])
			ds = ds.concatenate(label_ds)
		data, labels = None, label_ds.labels
	else:
		print("Could not read file due to bad interface")
		return None
	return ds, data, labels

def write_data(out_int, destination, name, data, ds, p=None, save_input=True):
	global model, verbose
	if verbose > 1 : print("Writing results")
	if p is None:
		filepath = destination + '/' + name + '-train' + out_int.filetypes[0][1][1:]
	else:
		filepath = destination + '/' + name + '-predictions' + out_int.filetypes[0][1][1:]
	if out_int.type == "RAW":
		if p is None:
			print("No predictions used, no need to write raw data again")
			return
		result = DataObject(name="P", data=p, resolution=data[0].resolution/data[0].divide)
		out_int.write(filepath + 'x', data, result, map=model.params['process']['SCORE_MAP'],
						epoch_size=model.params['process']['EPOCH_SIZE'])
	elif out_int.type == "DATASET":
		if p is None and not save_input:
			return
		elif p is None:
			result = ds
		elif p is not None:
			result = Dataset(label_names=['P'], labels=p)
			if save_input : result = ds.concatenate(result)
		out_int.write(filepath, result)
	else:
		print("Could not write results due to bad interface")

def classify():
	global model, verbose
	print("-"*30)
	print("Sleep Ensemble Classification")
	print("Current model:", model.name)
	print("-"*30)
	print("Select input interface")
	in_int = select_interface(tags={'r'})
	if in_int is None:
		print("You didn't select an interface! Returning you to the main menu")
		return
	print("Select files to classify")
	print("Currently configured to accept", in_int.standard)
	filepaths = ask_filenames(filetypes=in_int.filetypes)
	if len(filepaths) == 0:
		print("You didn't select any files! Returning you to the main menu")
		return
	print("Select where classifications should go:")
	destination = ask_directory()
	if destination is None or not os.path.isdir(destination):
		print("You didn't select a destination! Returning you to the main menu")
		return
	print("Select output interface")
	out_int = select_interface(tags={'w'})
	if out_int is None:
		print("You didn't select an interface! Returning you to the main menu")
		return
	save_input = yes_no_loop("Do you also want to export processed data alongside classifications?")
	print("Identified", len(filepaths), "files to classify")
	if verbose == 1 : jobs = trange(len(filepaths))
	else : jobs = range(len(filepaths))
	for i in jobs:
		if verbose > 1 : print("Classifying", filepaths[i])
		ds, data, labels = read_data(in_int, filepaths[i], labels=False)
		p, y_hat = model.predict([ds.data])
		write_data(out_int, destination, ds.name, data, ds, p=p[0], save_input=save_input)
	print("Completed classification jobs!")

def train():
	global model, verbose
	print("-"*30)
	print("Sleep Ensemble Training")
	print("Current model:", model.name)
	print("-"*30)
	print("1. Create a new model from a protocol class")
	print("2. Train from the current model")
	print("Q. Return to main menu")
	print("-"*30)
	usr = input("Select an option: ").upper()
	if usr.isdigit():
		usr = int(usr)
		if usr == 1:
			change_protocol()
		elif usr == 2:
			warm_start = yes_no_loop("Do you want a warm start with this model?")
			model.classifier.set_warm_start(warm_start)
		else : print("Sorry, didn't quite catch that.")
	elif usr == "Q": return
	else : print("Sorry, didn't quite catch that.")
	print("Training from model:", model.name)
	print("Select input interface")
	in_int = select_interface(tags={'r'})
	if in_int is None:
		print("You didn't select an interface! Returning you to the main menu")
		return
	print("Select files to use for training")
	print("Currently configured to accept", in_int.standard)
	filepaths = ask_filenames(filetypes=in_int.filetypes)
	if len(filepaths) == 0:
		print("You didn't select any files! Returning you to the main menu")
		return
	print("Identified", len(filepaths), "files for training")
	ds, data, labels = [], [], []
	if verbose == 1 : jobs = trange(len(filepaths))
	else : jobs = range(len(filepaths))
	for i in jobs:
		ds_, _, _ = read_data(in_int, filepaths[i], labels=True)
		ds.append(ds_)
		data.append(ds_.data)
		labels.append(ds_.labels)
	print("Training model on", len(np.concatenate(data)), "samples")
	model.fit(data, labels)
	print("Completed Training!")
	save_input = yes_no_loop("Do you want to export processed training data?")
	if save_input:
		print("Select output interface")
		out_int = select_interface(tags={'w'})
		if out_int is None:
			print("You didn't select an interface! Returning you to the main menu")
			return
		print("Select where data should be saved")
		destination = ask_directory()
		if destination is None or not os.path.isdir(destination):
			print("You didn't select a destination! Returning you to the main menu")
			return
		for ds_ in ds:
			write_data(out_int, destination, ds_.name, ds_.data, ds_)
	print("Complete! You can export model from the main menu")

def validate():
	global model, verbose
	print("-"*30)
	print("Sleep Ensemble Validation")
	print("Current model:", model.name)
	print("-"*30)
	print("Select input interface")
	in_int = select_interface(tags={'r'})
	if in_int is None:
		print("You didn't select an interface! Returning you to the main menu")
		return
	print("Select files to use for validation:")
	filepaths = ask_filenames(filetypes=in_int.filetypes)
	if len(filepaths) == 0:
		print("You didn't select any files! Returning you to the main menu")
		return
	save_input = yes_no_loop("Do you want to export the results?")
	if save_input:
		print("Select where results should go:")
		destination = ask_directory()
		if destination is None or not os.path.isdir(destination):
			print("You didn't select a destination! Returning you to the main menu")
			return
	print("Identified", len(filepaths), "files for validation")
	req_train = yes_no_loop("Do you wish to train and validate via cross-validation (y) or just validate (n)?")
	ds, data, labels = [], [], []
	raw_data = []
	if verbose == 1 : jobs = trange(len(filepaths))
	else : jobs = range(len(filepaths))
	for i in jobs:
		ds_, data_, labels_ = read_data(in_int, filepaths[i], labels=True)
		ds.append(ds_)
		data.append(ds_.data)
		labels.append(ds_.labels)
		raw_data.append(data_)
	if req_train:
		p, Y_hat = model.cross_validate(data, labels)
	else:
		p, Y_hat = model.predict(data)
	if p[0].ndim != 1:
		p_ = []
		start = 0
		for i in range(len(data)):
			p_.append(p[start:start + len(data[i])])
			start = start + len(data[i])
		p = p_
	score = model.score(p, labels)
	p_overall = np.concatenate(p)
	labels_overall = np.concatenate(labels)
	print_report(p_overall.reshape(-1), labels_overall.reshape(-1))
	if save_input:
		print("Select output interface")
		out_int = select_interface(tags={'w'})
		if out_int is None:
			print("You didn't select an interface! Returning you to the main menu")
			return
		for i in range(len(p)):
			write_data(out_int, destination, ds[i].name + "-%.4f" % score[i], raw_data[i], ds[i], p=p[i])
	print("Completed validation")

def load():
	global model, verbose
	print("-"*30)
	print("Load Model")
	print("Current model:", model.name)
	print("-"*30)
	protocol = change_protocol()
	if not protocol:
		print("You didn't select a protocol. Returning you to the main menu")
		return
	print("Selected", protocol.__name__, "protocol")
	model = protocol(verbose=model.verbose)
	file = ask_filename(title="Choose Model File", filetypes=[("Joblib", "*.joblib")])
	if not file:
		print("You didn't select a file! Returning you to the main menu")
		return
	print("Selected", Path(file).stem)
	confirm = yes_no_loop("Are you sure? Only load .joblib files from trusted sources")
	if not confirm:
		print("Cancelling import. Returning you to the main menu")
		return
	model.classifier = joblib.load(file)
	print("Loaded", model.name)

def export():
	global model, verbose
	print("-"*30)
	print("Export Model")
	print("Current model:", model.name)
	print("-"*30)
	name_change = yes_no_loop("Do you want to change the name of the model?")
	if name_change:
		model.name = input("Provide a name:")
	file = save_file(title="Save Model File", defaultextension=".joblib",
					filetypes=[("Joblib", "*.joblib")])
	if not file:
		print("Cancelling save. Returning you to the main menu")
		return
	print("Exporting", model.name, "to", Path(file).stem)
	joblib.dump(model.classifier, file, compress=5)
	print("Exported", model.name, "to", Path(file).stem)

def select_interface(tags={'r', 'w'}):
	global model, verbose
	while True:
		print("-"*30)
		print("Select File Interface")
		print("-"*30)
		accept = []
		for i in range(len(interfaces)):
			if tags.issubset(interfaces[i].tags):
				accept.append(interfaces[i])
		for i in range(len(accept)):
			print(str(i+1)+".", accept[i].name + ":", accept[i].standard, accept[i].tags)
		print("Q. Return to main menu")
		print("-"*30)
		usr = input("Select new file interface: ").upper()
		if usr.isdigit():
			usr = int(usr)
			if usr > 0 and usr <= len(accept):
				interface = accept[usr-1]
				print("Changed interface to", interface.name)
				return interface
			else : print("Sorry, didn't quite catch that.")
		elif usr == "Q": return
		else : print("Sorry, didn't quite catch that.")

def change_verbosity():
	global model, verbose
	while True:
		print("-"*30)
		print("Change Verbosity")
		print("Current verbosity:", verbose)
		print("-"*30)
		print("0. Silent (other than application messages)")
		print("1. Quiet - File level - Quiet")
		print("2. Quiet - File level")
		print("3. Soft - Model level")
		print("4. Soft - Classifier level - Quiet")
		print("5. Normal - Classifier level - Soft")
		print("6. Normal - Classifier level - Normal")
		print("7. Loud - Classifier level - Loud")
		print("Q. Return to main menu")
		print("-"*30)
		usr = input("Select an option: ").upper()
		if usr.isdigit():
			usr = int(usr)
			if usr >= 0 and usr <= 7:
				verbose = usr
				model.set_verbose(verbose - 1)
				print("Changed verbosity to", usr)
				return
			else : print("Sorry, didn't quite catch that.")
		elif usr == "Q": return
		else : print("Sorry, didn't quite catch that.")

def update_params():
	global model, verbose
	print("-"*30)
	print("Update Parameters")
	print("-"*30)
	pprint(model.params)
	print("-"*30)
	change = yes_no_loop("Do you wish to change parameter file?")
	if change:
		file = ask_filename(title="Choose Parameter File",
						filetypes=[("Python files", "*.py")])
		if not file:
			print("You didn't select a file! Returning you to the main menu")
			return
		print("Selected", Path(file).stem)
		confirm = yes_no_loop("Are you sure? Only load .py files from trusted sources")
		if not confirm:
			print("Cancelling import. Returning you to the main menu")
			return
		params = import_module('params', file)
		if not isinstance(params, dict):
			print("Imported something that isn't in a parameter dictionary! Returning you to the main menu")
			return
		print("Found the following:")
		pprint(params)
		finalize = yes_no_loop("Confirm update?")
		if finalize:
			model.params = params
			print("Updated parameters")
	print("Returning you to the main menu")

def main():
	print("Starting...")
	model.set_verbose(verbose - 1)
	while True:
		time.sleep(0.5)
		displayMenu()
		cmd = input("What would you like to do? ").upper()
		if cmd == "1":
			classify()
		elif cmd == "2":
			train()
		elif cmd == "3":
			validate()
		elif cmd == "4":
			load()
		elif cmd == "5":
			export()
		elif cmd == "6":
			change_verbosity()
		elif cmd == "7":
			update_params()
		elif cmd == "Q":
			print("Goodbye!")
			exit()
		else:
			print("Sorry, didn't quite catch that.")



"""Classifier and TimeSeriesClassifier"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

from abc import ABC, abstractmethod
import numpy as np
from tqdm import trange
from copy import deepcopy
from sklearn.utils.validation import check_is_fitted
from sklearn import metrics

from sleepens.utils import create_random_state, check_XY, one_hot

def check_estimator(estimator):
	"""
	Checks that `estimator` is a subclass of
	Classifier. Raises an error if
	it is not.

	Parameters
	----------
	estimator : object
		Object to check.
	"""
	if not issubclass(type(estimator), Classifier):
		raise ValueError("Object is not a Classifier")

def check_timeseries_estimator(estimator):
	"""
	Checks that `estimator` is a subclass of
	TimeSeriesClassifier. Raises an error if
	it is not.

	Parameters
	----------
	estimator : object
		Object to check.
	"""
	if not issubclass(type(estimator), TimeSeriesClassifier):
		raise ValueError("Object is not a TimeSeriesClassifier")

class Classifier(ABC):
	"""
	Base Classifier.

	Parameters
	----------
	warm_start : bool, default=False
		Determines warm starting to allow training to pick
		up from previous training sessions.

	random_state : None or int or RandomState, default=None
		Initial seed for the RandomState. If `random_state` is None,
		return the RandomState singleton. If `random_state` is an int,
		return a RandomState with the seed set to the int.
		If seed is a RandomState, return that RandomState.

	verbose : int, default=0
		Verbosity of estimator; higher values result in
		more verbose output.

	Attributes
	----------
	n_classes_ : int
		Number of classes.

	n_features_ : int
		Number of features.
	"""
	def __init__(self, warm_start=False,
				random_state=None, verbose=0):
		self.warm_start = warm_start
		self.random_state = create_random_state(random_state)
		self.verbose = verbose

	@abstractmethod
	def fit(self, X, Y):
		"""
		Train the classifier on the given data and labels.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Training data.

		Y : array-like, shape=(n_samples,)
			Target labels as integers.

		Returns
		-------
		self : Classifier
			Fitted classifier.
		"""
		raise NotImplementedError("No fit function implemented")

	def predict(self, X):
		"""
		Predict on the given data.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Data to predict.

		Returns
		-------
		Y_hat : array-like, shape=(n_samples,)
			Predictions.
		"""
		return np.argmax(self.predict_proba(X), axis=1)

	@abstractmethod
	def predict_proba(self, X):
		"""
		Return the prediction probabilities on the given data.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Data to predict.

		Returns
		-------
		Y_hat : array-like, shape=(n_samples, n_classes)
			Probability predictions.
		"""
		raise NotImplementedError("No predict_proba function implemented")

	def predict_log_proba(self, X):
		"""
		Return the prediction log probabilities on the given data.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Data to predict.

		Returns
		-------
		Y_hat : array-like, shape=(n_samples, n_classes)
			Log probability predictions.
		"""
		return np.log(self.predict_proba(X))

	def feature_importance(self, X, Y, repeats=10, sort=True):
		"""
		Return the importances of features through Permutation
		Feature Importance (PFI). In PFI, the fitted classifier
		predicts over data where one feature has been permuted randomly.
		The resulting change in mean square error provides a measure of that feature's
		importance.

		Take care when working with data that contains correlated features.
		The results of PFI may not represent true feature importances in such cases,
		due to either permutations resulting in impossible data or a reduction in
		apparent feature importance due to the sharing of utility among correlated features.

		If conducted with the training data, PFI shows the most salient features
		used by the classifier. If conducted with unseen data, PFI describes
		the most valuable features for predictions.

		Remember that PFI relates to the specific classifier in question.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Data.

		Y : array-like, shape=(n_samples,)
			Target labels as integers.

		repeats : int, default=10
			Number of times to repeat permutations. Importances
			the average across repetitions.

		sort : bool, default=True
			Determine if the feature importances should be sorted.

		Returns
		-------
		importances : list of tuple (feature, importance)
			List of each feature and the corresponding importance.
			The feature is labelled numerically in the order they appear
			in the data. Importance is the change in the score
			from permuting that feature.
		"""
		if not self._is_fitted():
			raise RuntimeError("Model has not been initialized")
		X, Y = check_XY(X=X, Y=Y)
		if self.verbose > 1 : print("Calculating feature importances")
		Y_one_hot = one_hot(Y)
		p = self.predict_proba(X)
		error = np.mean(np.square(Y_one_hot - p))
		importances = np.zeros((repeats, X_.shape[1]))
		if self.verbose == 2 : trn = trange(repeats)
		else : trn = range(repeats)
		for n in trn:
			if self.verbose == 3 : print("Repeat", n+1)
			if self.verbose == 3 : trf = trange(X.shape[1])
			else : trf = range(X.shape[1])
			for f in trf:
				if self.verbose > 3 : print("Permuting feature", f+1)
				self.set_verbose(self.verbose - 3)
				X_p = deepcopy(X)
				self.random_state.shuffle(X_p[:,f])
				p_ = self.predict_proba(X_p)
				error_ = np.mean(np.square(Y_one_hot - p_))
				importances[n, f] = (error_ / error)
				self.set_verbose(self.verbose + 3)
		mu = np.mean(importances, axis=0)
		sigma = np.std(importances, axis=0)
		importances = [(i, mu[i], sigma[i]) for i in range(X_.shape[1])]
		if sort:
			return sorted(importances, key=lambda x: x[1], reverse=True)
		return importances

	def score(self, Y_hat, Y):
		"""
		Calculate the accuracy of the given predictions compared
		to the target labels.

		Parameters
		----------
		Y_hat : array-like, shape=(n_samples,)
			Classifier predictions.

		Y : array-like, shape=(n_samples,)
			Target labels.

		Returns
		-------
		score : float, [0,1)
			The calculated score. If `metric` is None,
			returns 0.
		"""
		return metrics.accuracy_score(Y, Y_hat)

	def set_verbose(self, verbose):
		"""
		Set the verbosity of the Classifier.

		Parameters
		----------
		verbose : int, default=0
			Determines the verbosity of cross-validation.
			Higher verbose levels result in more output logged.
		"""
		self.verbose = verbose

	def set_random_state(self, random_state):
		"""
		Set the RandomState of the Classifier.

		Parameters
		----------
		random_state : None or int or RandomState, default=None
			Initial seed for the RandomState. If `random_state` is None,
			return the RandomState singleton. If `random_state` is an int,
			return a RandomState with the seed set to the int.
			If `random_state` is a RandomState, return that RandomState.
		"""
		self.random_state = create_random_state(random_state)

	def set_warm_start(self, warm_start):
		"""
		Set the status of `warm_start` of the Classifier.

		Parameters
		----------
		warm_start : bool
			Determines warm starting to allow training to pick
			up from previous training sessions.
		"""
		self.warm_start = warm_start

	def _initialize(self):
		"""
		Initialize the parameters of the classifier.
		"""
		return

	def _fit_setup(self, X, Y):
		"""
		Set up the fit process.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Training data.

		Y : array-like, shape=(n_samples,)
			Target labels as integers.

		Returns
		-------
		X : ndarray, shape=(n_samples, n_features)
			Training data.

		Y : ndarray, shape=(n_samples,)
			Target labels as integers.
		"""
		X, Y = check_XY(X=X, Y=Y)
		self._initialize()
		self._class_features(len(np.unique(Y)), X.shape[-1])
		if self.verbose > 1 : print("Training model")
		return X, Y

	def _predict_setup(self, X):
		"""
		Set up the predict process.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Data to predict.

		Returns
		-------
		X : ndarray, shape=(n_samples, n_features)
			Data to predict.
		"""
		if not self._is_fitted():
			raise RuntimeError("Model has not been fitted")
		X, _ = check_XY(X=X)
		if X.shape[1] != self.n_features_:
			raise ValueError("Model expects %d features" % self.n_features_,
								"but input has %d features" % X.shape[1])
		if self.verbose > 1 : print("Predicting %d samples." % X.shape[0])
		return X

	def _is_fitted(self, attributes=["n_classes_","n_features_"]):
		"""
		Returns if the Classifier has been trained and is
		ready to predict new data.

		Parameters
		----------
		attributes : list of string, default=["n_classes_","n_features_"]
			List of instance attributes to check existence of as proof
			that the Classifier has been fitted. Therefore, these attributes
			should only be created during the fit process.

		Returns
		-------
		fitted : bool
			True if the Classifier is fitted, False otherwise.
		"""
		try:
			check_is_fitted(self, attributes=attributes)
			return True
		except : return False

	def _class_features(self, n_classes, n_features):
		"""
		Set up the number of classes and features this
		model is designed to work with.

		If the Classifier has `warm_start` as True and
		already has `n_classes_` or `n_features_` set,
		the provided arguments must match.

		Parameters
		----------
		n_classes : int
			Number of classes.

		n_features : int
			Number of features.
		"""
		if not hasattr(self, 'n_classes_') or not self.warm_start : self.n_classes_ = n_classes
		elif self.warm_start and self.n_classes_ < n_classes:
			raise ValueError("Class mismatch: Model was trained on", self.n_classes_,
							"classes but input has", n_classes, "classes")
		if not hasattr(self, 'n_features_') or not self.warm_start : self.n_features_ = n_features
		elif self.warm_start and self.n_features_ != n_features:
			raise ValueError("Feature mismatch: Model has", self.n_features_,
							"features but input has", n_features, "features")

class TimeSeriesClassifier(Classifier):
	"""
	Base Time Series Classifier.

	Parameters
	----------
	warm_start : bool, default=False
		Determines warm starting to allow training to pick
		up from previous training sessions.

	random_state : None or int or RandomState, default=None
		Initial seed for the RandomState. If `random_state` is None,
		return the RandomState singleton. If `random_state` is an int,
		return a RandomState with the seed set to the int.
		If seed is a RandomState, return that RandomState.

	verbose : int, default=0
		Verbosity of estimator; higher values result in
		more verbose output.

	Attributes
	----------
	n_classes_ : int
		Number of classes.

	n_features_ : int
		Number of features.
	"""
	def __init__(self, warm_start=False,
				random_state=None, verbose=0):
		Classifier.__init__(self, warm_start=warm_start,
				random_state=random_state, verbose=verbose)

	@abstractmethod
	def fit(self, X, Y):
		"""
		Train the classifier on the given data and labels.

		Parameters
		----------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Training data.

		Y : list of ndarray, shape=(s_series, n_samples)
			Target labels as integers.

		Returns
		-------
		self : Classifier
			Fitted classifier.
		"""
		raise NotImplementedError("No fit function implemented")

	def predict(self, X):
		"""
		Predict on the given data.

		Parameters
		----------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Data to predict.

		Returns
		-------
		Y_hat : list of ndarray, shape=(s_series, n_samples)
			Predictions.
		"""
		return [np.argmax(y, axis=1) for y in self.predict_proba(X)]

	@abstractmethod
	def predict_proba(self, X):
		"""
		Return the prediction probabilities on the given data.

		Parameters
		----------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Data to predict.

		Returns
		-------
		Y_hat : list of ndarray, shape=(s_series, n_samples, n_classes)
			Probability predictions.
		"""
		raise NotImplementedError("No predict_proba function implemented")

	def predict_log_proba(self, X):
		"""
		Return the prediction log probabilities on the given data.

		Parameters
		----------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Data to predict.

		Returns
		-------
		Y_hat : list of ndarray, shape=(s_series, n_samples, n_classes)
			Log probability predictions.
		"""
		return [np.log(y) for y in self.predict_proba(X)]

	def feature_importance(self, X, Y, repeats=10, sort=True):
		"""
		Return the importances of features through Permutation
		Feature Importance (PFI). In PFI, the fitted classifier
		predicts over data where one feature has been permuted randomly.
		The resulting change in mean squared error provides a measure of
		that feature's importance.

		Take care when working with data that contains correlated features.
		The results of PFI may not represent true feature importances in such cases,
		due to either permutations resulting in impossible data or a reduction in
		apparent feature importance due to the sharing of utility among correlated features.

		If conducted with the training data, PFI shows the most salient features
		used by the classifier. If conducted with unseen data, PFI describes
		the most valuable features for predictions.

		Remember that PFI relates to the specific classifier in question.

		Parameters
		----------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Data.

		Y : list of ndarray, shape=(n_series, n_samples,)
			Target labels as integers.

		repeats : int, default=10
			Number of times to repeat permutations. Importances
			the average across repetitions.

		sort : bool, default=True
			Determine if the feature importances should be sorted.

		Returns
		-------
		importances : list of tuple (feature, importance)
			List of each feature and the corresponding importance.
			The feature is labelled numerically in the order they appear
			in the data. Importance is the change in the score
			from permuting that feature.
		"""
		if not self._is_fitted():
			raise RuntimeError("Model has not been initialized")
		for x, y in zip(X, Y) : check_XY(X=x, Y=y)
		try : X_, Y_ = np.concatenate(X), np.concatenate(Y)
		except : raise ValueError("Inputs have different number of features")
		if self.verbose > 1 : print("Calculating feature importances")
		Y_one_hot = np.concatenate([one_hot(y) for y in Y])
		p = np.concatenate(self.predict_proba(X))
		error = np.mean(np.square(Y_one_hot - p))
		importances = np.zeros((repeats, X_.shape[1]))
		if self.verbose == 2 : trn = trange(repeats)
		else : trn = range(repeats)
		for n in trn:
			if self.verbose == 3 : print("Repeat", n+1)
			if self.verbose == 3 : trf = trange(X_.shape[1])
			else : trf = range(X_.shape[1])
			for f in trf:
				if self.verbose > 3 : print("Permuting feature", f+1)
				self.set_verbose(self.verbose - 3)
				X_p = [deepcopy(x) for x in X]
				for x in X_p : self.random_state.shuffle(x[:,f])
				p_ = np.concatenate(self.predict_proba(X_p))
				error_ = np.mean(np.square(Y_one_hot - p_))
				importances[n, f] = (error_ / error)
				self.set_verbose(self.verbose + 3)
		mu = np.mean(importances, axis=0)
		sigma = np.std(importances, axis=0)
		importances = [(i, mu[i], sigma[i]) for i in range(X_.shape[1])]
		if sort:
			return sorted(importances, key=lambda x: x[1], reverse=True)
		return importances

	def score(self, Y_hat, Y):
		"""
		Calculate the accuracy of the given predictions compared
		to the target labels.

		Parameters
		----------
		Y_hat : list of ndarray, shape=(n_series, n_samples,)
			Classifier predictions.

		Y : list of ndarray, shape=(n_series, n_samples,)
			Target labels.

		Returns
		-------
		score : float, [0,1)
			The calculated score. If `metric` is None,
			returns 0.
		"""
		scores = [metrics.accuracy_score(y, y_hat) for y_hat, y in zip(Y_hat, Y)]
		lengths = [len(y_hat) for y_hat in Y_hat]
		return np.average(scores, weights=lengths)

	def _fit_setup(self, X, Y):
		"""
		Set up the fit process.

		Parameters
		----------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Training data.

		Y : list of ndarray, shape=(s_series, n_samples)
			Target labels as integers.

		Returns
		-------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Training data.

		Y : list of ndarray, shape=(s_series, n_samples)
			Target labels as integers.
		"""
		for x, y in zip(X, Y) : check_XY(X=x, Y=y)
		try : X_, Y_ = np.concatenate(X), np.concatenate(Y)
		except : raise ValueError("Inputs have different number of features")
		self._initialize()
		self._class_features(len(np.unique(Y_)), X_.shape[-1])
		if self.verbose > 1 : print("Training model")
		return X, Y

	def _predict_setup(self, X):
		"""
		Set up the predict process.

		Parameters
		----------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Data to predict.

		Returns
		-------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Data to predict.
		"""
		if not self._is_fitted():
			raise RuntimeError("Model has not been fitted")
		for x in X : check_XY(X=x)
		try : X_ = np.concatenate(X)
		except : raise ValueError("Inputs have different number of features")
		if X_.shape[1] != self.n_features_:
			raise ValueError("Model expects %d features" % self.n_features_,
								"but input has %d features" % X_.shape[1])
		if self.verbose > 1 : print("Predicting %d samples." % X_.shape[0])
		return X



import numpy as np
from tqdm import trange
from copy import deepcopy

from sleepens.ml import check_estimator
from sleepens.ml._base_model import TimeSeriesClassifier
from sleepens.ml.models import GradientBoostingClassifier
from sleepens.utils import time_array

default_estimators = [GradientBoostingClassifier() for i in range(10)]

class TimeSeriesEnsemble(TimeSeriesClassifier):
	"""
	A Time Series Ensemble is an Ensemble Classifier that
	uses an array of estimators to classify time series data
	with a rolling window.

	Each ensemble member views the same window of the time series
	but predicts at a given timepoint within the window. All predictions
	from visited ensemble members are then averaged to produce the overall
	prediction.

	The Time Series Ensemble provides two main advantages:
	1. A rolling window style of classification that does not truncate
		data on either end.
	2. A rolling window that is tapers to twice the size of the given window.
	Finally, the Time Series Ensemble is implemented as a bi-directional
	Classifier, evaluating with both past and future timepoints.

	Parameters
	----------
	estimators : list of Classifier, default=[10*GradientBoostingClassifier]
		The ensemble array. The number of classifiers is assumed
		as the window size. The i-th Classifier will predict for
		the i-th timepoint within the moving window.

	warm_start : bool, default=False
		Determines warm starting to allow training to pick
		up from previous training sessions.

	random_state : None or int or RandomState, default=None
		Initial seed for the RandomState. If `random_state` is None,
		return the RandomState singleton. If `random_state` is an int,
		return a RandomState with the seed set to the int.
		If seed is a RandomState, return that RandomState.

	verbose : int, default=0
		Verbosity of estimator; higher values result in
		more verbose output.

	Attributes
	----------
	n_classes_ : int
		Number of classes.

	n_features_ : int
		Number of features.
	"""
	def __init__(self, estimators=default_estimators,
					warm_start=False, random_state=None, verbose=0):
		TimeSeriesClassifier.__init__(self, warm_start=warm_start,
							random_state=random_state, verbose=verbose)
		for e in estimators:
			check_estimator(e)
		self.n = len(estimators)
		self.base_estimators = estimators
		self.set_warm_start(warm_start)
		self.set_verbose(verbose)
		self.set_random_state(random_state)

	def _initialize(self):
		"""
		Initialize the parameters of the classifier.
		"""
		if self._is_fitted():
			for e in self.estimators_ : e.set_warm_start(self.warm_start)
		else:
			self.estimators_ = deepcopy(self.base_estimators)

	def fit(self, X, Y):
		"""
		Train the classifier on the given data and labels.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Training data.

		Y : array-like, shape=(n_samples,)
			Target labels as integers.

		Returns
		-------
		self : Classifier
			Fitted classifier.
		"""
		X, Y = self._fit_setup(X, Y)
		X_a, Y_a = [time_array(x, self.n) for x in X], [time_array(y, self.n) for y in Y]
		X_a, Y_a = np.concatenate(X_a), np.concatenate(Y_a)
		if self.verbose == 1 or self.verbose == 2 : stages = trange(self.n)
		else : stages = range(self.n)
		for s in stages:
			if self.verbose > 2 : print("Fitting estimator", s)
			self.estimators_[s].fit(X_a, Y_a[:,s])
		if self.verbose > 1 : print("Completed training")
		return self

	def predict_proba(self, X):
		"""
		Return the prediction probabilities on the given data.

		Parameters
		----------
		X : list of ndarray, shape=(n_series, n_samples, n_features)
			Data to predict.

		Returns
		-------
		Y_hat : list of ndarray, shape=(s_series, n_samples, n_classes)
			Probability predictions.
		"""
		X = self._predict_setup(X)
		X_a = [time_array(x, self.n) for x in X]
		Y_hat = [np.zeros((len(x), self.n_classes_)) for x in X]
		count = [np.zeros((len(x),), dtype=int) for x in X]
		if self.verbose == 1 or self.verbose == 2 : stages = trange(self.n)
		else : stages = range(self.n)
		for s in stages:
			if self.verbose > 2 : print("Predicting on estimator", s)
			for i in range(len(X)):
				Y_hat[i][s:s+len(X_a[i])] += self.estimators_[s].predict_proba(X_a[i])
				count[i][s:s+len(X_a[i])] += 1
		for i in range(len(Y_hat)) : Y_hat[i] /= count[i].reshape(-1,1)
		if self.verbose > 1 : print("Completed predicting")
		return Y_hat

	def set_verbose(self, verbose):
		"""
		Set the verbosity of the Classifier.
		Subsidiary Classifiers have their verbosity
		suppressed by one compared to this Classifier.

		Parameters
		----------
		verbose : int, default=0
			Determines the verbosity of cross-validation.
			Higher verbose levels result in more output logged.
		"""
		TimeSeriesClassifier.set_verbose(self, verbose)
		estimators = self.estimators_ if hasattr(self, 'estimators_') else self.base_estimators
		for e in estimators:
			e.set_verbose(verbose-1)

	def set_random_state(self, random_state):
		"""
		Set the RandomState of the Classifier.
		Subsidiary Classifiers set their RandomState
		based on this RandomState with a seed selected
		from 0 to 2**16.

		Parameters
		----------
		random_state : None or int or RandomState, default=None
			Initial seed for the RandomState. If `random_state` is None,
			return the RandomState singleton. If `random_state` is an int,
			return a RandomState with the seed set to the int.
			If `random_state` is a RandomState, return that RandomState.
		"""
		TimeSeriesClassifier.set_random_state(self, random_state)
		estimators = self.estimators_ if hasattr(self, 'estimators_') else self.base_estimators
		for e in estimators:
			e.set_random_state(self.random_state.randint(0, 2**16))

	def set_warm_start(self, warm_start):
		"""
		Set the status of `warm_start` of the Classifier.

		Parameters
		----------
		warm_start : bool
			Determines warm starting to allow training to pick
			up from previous training sessions.
		"""
		TimeSeriesClassifier.set_warm_start(self, warm_start)
		estimators = self.estimators_ if hasattr(self, 'estimators_') else self.base_estimators
		for e in estimators:
			e.set_warm_start(warm_start)

	def _is_fitted(self):
		"""
		Returns if the Classifier has been trained and is
		ready to predict new data.

		Returns
		-------
		fitted : bool
			True if the Classifier is fitted, False otherwise.
		"""
		attributes = ["estimators_","n_classes_","n_features_"]
		return TimeSeriesClassifier._is_fitted(self, attributes=attributes)



"""Ratio Processor"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.io import Dataset
from sleepens.utils.data import signal as s

def ratio(ds, name=None, ratios={}):
	"""
	Process the epoched power spectral density
	and compute ratio between two bands.

	Parameters
	----------


	Returns
	-------
	ratios : ndarray, shape=(n_epochs,)
		List of ratios at each epoch.
	"""
	data = ds.data
	features = ds.features.tolist()
	results = []
	ratio_names = list(ratios.keys())
	for ratio in ratio_names:
		num = ratios[ratio][0]
		den = ratios[ratio][1]
		numerator = data[:, features.index(num)].flatten()
		denom = data[:, features.index(den)].flatten()
		results.append(np.divide(numerator, denom))
	results = np.array(results).T
	return Dataset(name=name, features=ratio_names, data=results)



"""Spectral Band Processor"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.io import Dataset
from sleepens.utils.data import signal as s

def spectral_band(ds, name=None, bands=None, merge='sum'):
	"""
	Process the epoched power spectral density
	and compute power at a specific frequency
	band.

	Parameters
	----------
	f : ndarray
		List of arrays of frequencies for each epoch.

	Pxx : ndarray
		List of power spectral densities of `data` at
		each epoch.

	params : dict
		Dictionary of all band parameters:
		 - BANDS :  Dictionary of tuple intervals for bands.
		 - BAND_MERGE : Method to merge power spectral densities
		 				within bands. Select from
						{'mean', 'max', 'min', 'sum'}.

	Returns
	-------

	"""
	f, Pxx = ds.features, ds.data
	if bands is None : bands = {'ALL': (np.min(f), np.max(f))}
	names = list(bands.keys())
	data = []
	for e in range(min([len(f), len(Pxx)])):
		band = s.compute_bands(f[e], Pxx[e], bands, merge)
		band = [band[k] for k in names]
		data.append(band)
	data = np.array(data)
	return Dataset(name=name, features=names, data=data)



"""Percentile Mean Processor"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.io import DataObject, Dataset
from sleepens.utils.data import signal as s
from sleepens.utils import calculate_epochs, get_epoch

def rms(dataobject, epoch_size, name=None):
	"""
	Process the data in `dataobject`,
	divide the data into epochs of `epoch_size`, and
	compute the root-mean-square at each epoch.

	Parameters
	----------
	dataobject : DataObject
		DataObject for processing.

	epoch_size : int
		Number of seconds in each epoch.

	Returns
	-------
	ds = Dataset
		Dataset with root-mean-square at each epoch.
		Feature is named "RMS"
	"""
	rms = []
	epoch_len, n_epochs = calculate_epochs(dataobject.data, dataobject.resolution, epoch_size)
	for i in range(n_epochs):
		epoch = get_epoch(dataobject.data, i, epoch_len)
		rms.append(s.rms(epoch))
	rms = np.array(rms).reshape(-1,1)
	return Dataset(name=name, features=["RMS"], data=rms)



from ._epoched_variance import epoched_variance
from ._fft import fft
from ._percentile_mean import percentile_mean
from ._rms import rms

__all__ = [
			'epoched_variance',
			'fft',
			'percentile_mean',
			'rms'
		]



default_map = { 'AW': 0, 'QW': 1, 'NR': 2, 'R': 3 }

from ._min_rem import MinREM
from ._rem_drop import REMDrop
from ._rem_sensitivity import REMSensitivity
from ._transition_fix import TransitionFix
from ._wake_max import WakeMax
from ._wake_sensitivity import WakeSensitivity
from ._wake_to_rem import WakeToREM

__all__ = [
			'MinRem',
			'RemDrop',
			'REMSensitivity',
			'TransitionFix',
			'WakeMax',
			'WakeSensitivity',
			'WakeToREM',
		]



"""MinREM Fix"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.postprocess import default_map

def MinREM(Y_hat, p, map=default_map, min_rem=3):
	"""
	Minimum REM length addon. Force REM episodes
	to be of minimum length.

	Fix is conducted by a backwards pass through predictions
	and brute repairing predictions to satisfy the minimum
	episode length requirement.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples, n_classes)
		The raw prediction probabilities.

	p : array-like, shape=(n_samples,)
		The predictions to process. If no Addon
		processing was done prior to this, `p`
		corresponds with Y_hat.

	map : dict
		Mapping the label values to some
		set of integers.

	min_rem : int, default=3
		Minimum length of a REM episode

	Returns
	-------
	p : array-like, shape=(n_samples,)
		The post-processed predictions.
	"""
	in_rem = 0
	for i in range(len(p)-1,-1,-1):
		if p[i] == map['R'] : in_rem += 1
		elif in_rem > 0 and in_rem < min_rem:
			p[i] = map['R']
			in_rem += 1
		else : in_rem = 0
	return p



from sleepens.utils.data import normalize as n

params = {
	'reader': {
		'EEG_NAME': 'EEG',
		'NECK_NAME': 'NECK',
		'MASS_NAME': 'MASS',
		'SCORE_NAME': 'SCORE',
		'SCORE_MAP': { 'AW': 0, 'QW': 1, 'NR': 2, 'R': 3 },
	},
	'process': {
		'EPOCH_SIZE': 5,
		'EEG_NAME': 'EEG',
		'NECK_NAME': 'NECK',
		'MASS_NAME': 'MASS',
		'SCORE_NAME': 'SCORE',
		'SCORE_MAP': { 0: 'AW', 1: 'QW', 2: 'NR', 3: 'R' },
		'INVERSE_MAP': { 'AW': 0, 'QW': 1, 'NR': 2, 'R': 3 },
		'EEG_FFT': {
						'NPERSEG': 0.75,
						'NOVERLAP': 0.5,
						'DETREND': 'constant',
					},
		'EMG_FFT': {
						'NPERSEG': 0.75,
						'NOVERLAP': 0.5,
						'DETREND': 'constant',
					},
		'PRCTILE': 95,
		'EPOCHED_VAR': {
						'SUBEPOCH_SIZE': 10,
						'THRESHOLD': 'median',
						'MERGE': 'sum'
					},
		'BANDS': {
					'DELTA' : (0.5, 4),
					'THETA' : (7, 10),
					'ALPHA'  : (11, 15),
					'BETA' : (15, 40),
				},
		'BAND_MERGE': 'sum',
		'RATIOS': {
					'THETA/DELTA': ('THETA','DELTA'),
					'ALPHA/DELTA':('ALPHA','DELTA'),
					'BETA/DELTA':('BETA','DELTA'),
					'ALPHA/THETA':('ALPHA','THETA'),
					'BETA/THETA':('BETA','THETA'),
					'BETA/ALPHA':('BETA','ALPHA')
				},
		'EMG_RMS_MERGE': 'sum',
		'EMG_PRCT_MERGE': 'sum',
		'EMG_TWITCH_MERGE': 'sum',
		'EMG_ENTROPY_MERGE': 'mean',
		'TRANSFORM': [{'FUNCTION':n.log_normalize, 'ARGS':['median', (-5, 5), 0]}]
	},
	'classifier': {
		'gbc': {
				'n_estimators': 250,
				'max_features': 'sqrt',
				'learning_rate': 0.05,
				'subsample': 0.75,
				'max_depth': 5
			}
	},
	'transition': [[1, 1, 1, 0],
					[1, 1, 1, 0],
					[1, 1, 1, 1],
					[1, 1, 0, 2]]
}



"""Smoothing Functions"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

def moving_average(data, window, weight=False, axis=0):
	"""
	Compute a moving average of the data.
	Moving average is calculated as average
	of the last `window` samples.

	Parameters
	----------
	data : ndarray
		Data.

	window : int
		Size of the moving average window.

	weight : bool, default=False
		Compute a weighted moving average.

	axis : int, default=0
		Axis to smooth along.

	Returns
	-------
	ma : ndarray
		Moving average of the data.
	"""
	result = []
	for i in range(len(data)):
		if i - window < 0:
			start = 0
		else:
			start = i - window + 1
		if weight:
			weights = np.arange(i-start+1) + 1
			sum = np.sum(data[start:i+1] * weights, axis=axis)
			avg = sum / np.sum(weights)
		else:
			sum = np.sum(data[start:i+1], axis=axis)
			avg = sum / window
		result.append(avg)
	return np.array(result)

def hull_moving_average(data, window, weight=False, axis=0):
	"""
	Compute a Hull moving average of the data.
	This is less susceptible to lag effects.
	Calculated as the moving average of
	Moving average is calculated as average
	of the last `window` samples.

	Parameters
	----------
	data : ndarray
		Data.

	window : int
		Size of the moving average window.

	weight : bool, default=False
		Compute a weighted moving average.

	axis : int, default=0
		Axis to smooth along.

	Returns
	-------
	hma : ndarray
		Hull moving average of the data.
	"""
	period = int(np.sqrt(window))
	half = int(window / 2)
	if half == 0 : half = 1
	halfaverage = moving_average(data, half, weight, axis=axis)
	fullaverage = moving_average(data, window, weight, axis=axis)
	intermediate = 2 * halfaverage - fullaverage
	result = moving_average(intermediate, period, weight, axis=axis)
	return result



from ._merge import merge
from ._ratio import ratio
from ._spectral_band import spectral_band
from ._spectral_entropy import spectral_entropy
from ._transform import transform

__all__ = [
			'merge',
			'ratio',
			'spectral_band',
			'spectral_entropy',
			'transform'
		]



"""Merge Processor"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.io import Dataset
from sleepens.utils import aggregate

def merge(ds, name=None, feature='MERGE', method='mean', axis=1):
	"""
	Merge features into a single feature
	by the given method.

	Parameters
	----------
	features : list
		List of 1-D feature ndarrays.

	method : {'mean', 'median', 'max', 'min', 'sum'}, default='mean'
		Method to compute the summary statistic of the band.

	Returns
	-------
	feature : ndarray
		Merged feature data.
	"""
	data = ds.data
	data = aggregate(data, method, axis=axis)
	data = data.reshape(-1,1)
	return Dataset(name=name, features=[feature], data=data)



"""Percentile Mean Processor"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.io import DataObject, Dataset
from sleepens.utils.data import signal as s
from sleepens.utils import aggregate, determine_threshold, calculate_epochs, get_epoch

def epoched_variance(dataobject, epoch_size, name=None, sub_epoch_size=10,
				threshold='median', merge='sum'):
	"""
	Process the data in `dataobject`,
	divide the data into epochs of `epoch_size`, and
	compute mean of data above the `k`-th percentile.

	Parameters
	----------
	dataobject : DataObject
		DataObject for processing.

	epoch_size : float
		Number of seconds in each epoch.

	sub_epoch_size : int
		Number of sub-epochs in an epoch.

	threshold : {'mean', 'median', '%<k>', 'float'}
		Threshold to count as 'activity'.
		Must be one of:
		 - 'mean' : Mean of sub-epochs.
		 - 'median' : Median of sub-epochs.
		 - '%<k>' : <k>-th percentile of sub-epochs.
		 			<k> is a non-negative integer.
		 - float : Constant threshold.

	merge : {'mean', 'median', 'max', 'min', 'sum', 'count'}
	Determines merge method.
	Must be one of:
	 - 'mean' : Mean of variances above threshold.
	 - 'median' : Median of variances above threshold.
	 - 'max' : Max of variances above threshold.
	 - 'min' : Min of variances above threshold.
	 - 'sum' : Sum of variances above threshold.
	 - 'count' : Count of variances above threshold.

	Returns
	-------
	ds = Dataset
		Dataset with epoched variance at each epoch.
		Feature is named "EPOCHED VARIANCE"
	"""
	activities = []
	epoch_len, n_epochs = calculate_epochs(dataobject.data, dataobject.resolution, epoch_size)
	for i in range(n_epochs):
		epoch = get_epoch(dataobject.data, i, epoch_len)
		subepoch_var = []
		subepoch_len = epoch_len // sub_epoch_size
		for j in range(sub_epoch_size):
			subepoch = epoch[j*subepoch_len:(j+1)*subepoch_len]
			subepoch_var.append(np.var(subepoch))
		activities.append(subepoch_var)
	activities = np.array(activities)
	shape = activities.shape
	threshold = determine_threshold(activities.flatten(), threshold)
	activities = list(activities.reshape(shape))
	for i in range(n_epochs):
		above = np.where(activities[i] >= threshold)
		activities[i] = aggregate(activities[i][above], merge)
	activities = np.array(activities).reshape(-1,1)
	ds = Dataset(name=name, features=["EPOCHED VARIANCE"], data=activities)
	return ds



"""Data Manipulation Functions"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np
from math import floor

from sleepens.utils import check_X

def time_array(X, n):
	"""
	Sequence time series data with a rolling window of size `n`.

	Parameters
	----------
	X : array-like, shape=(n_samples, n_features)
		Time series data.

	n : int
		Window size.

	Returns
	-------
	ta : array-like, shape=(n_samples-n+1, n_features*n)
		Time arrayed data.
	"""
	X = check_X(X)
	if X.ndim <= 1 : X = X.reshape(-1,1)
	if len(X) < n : raise ValueError("Length of array is smaller than n")
	end = len(X) - n + 1
	ta = X[:end]
	for i in range(1,n):
		ta = np.concatenate((ta, X[i:end+i]), axis=1)
	return ta

def separate_by_label(data, labels):
	"""
	Create a dictionary grouping data by labels
	in order.

	Parameters
	----------
	data : array-like, shape=(n_samples, n_features)
		Data.

	labels : array-like, shape=(n_samples,)
		Labels for the given `data`.

	Returns
	-------
	separated : dict
		Dictionary of categorized data.
	"""
	separated = {k: [] for k in labels}
	for d, l in zip(data, labels):
		separated[l].append(d)
	return separated

def one_hot(Y, cols=None):
	"""
	Convert `Y` into a one-hot encoding.

	Parameters
	----------
	Y : ndarray, shape=(n_samples,)
		Data to encode.

	cols : int or None, default=None
		Number of columns to encode.
		Must be at least the number of unique
		values in `Y`.

	Returns
	-------
	one_hot : ndarray, shape=(n_samples, n_cols)
		One-hot encoded data.
	"""
	if Y.shape[1] == cols and (set(Y.reshape(-1)) == set([0,1]) or \
			set(Y.reshape(-1)) == set([0.,1.])):
		return Y.astype(int)
	if Y.size > len(Y) : raise ValueError("Matrix is not 1D")
	Y_ = Y.reshape(-1)
	if cols is None : cols = Y_.max() + 1
	elif cols < len(np.unique(Y)) : raise ValueError("There are more classes than cols")
	if cols > 1:
		one_hot = np.zeros((len(Y), cols), dtype=int)
		one_hot[np.arange(len(Y)), Y_] = 1
		return one_hot
	else:
		Y = np.where(Y < cols, 0, 1)
		return Y.reshape(-1, 1)

def decode(Y):
	"""
	Decode one-hot encoded data
	into a 1-dimensional array.

	Parameters
	----------
	Y : ndarray, shape=(n_samples, n_cols)
		One-hot encoded data.

	Returns
	-------
	decode : ndarray, shape=(n_samples,)
		Decoded data.
	"""
	if Y.shape[0] == Y.size:
		return np.squeeze(Y)
	elif set(Y.reshape(-1)) == set([0,1]) or \
			set(Y.reshape(-1)) == set([0.,1.]):
		return np.argmax(Y, axis=1)
	else:
		raise ValueError("Y must be one-hot encoded data with ints.")

def calculate_epochs(data, resolution, epoch_size):
	"""
	Calculate the epoch length and number of epochs
	in the data based on `epoch_size`.

	Parameters
	----------
	data : array-like, shape=(n_samples, n_features)
		Data to be epoched.

	epoch_size : int
		Number of seconds in each epoch.

	Returns
	-------
	epoch_len : int
		Number of samples in each epoch.

	n_epochs : int
		Number of full epochs in the data.
	"""
	epoch_len = floor(epoch_size / resolution)
	n_epochs = floor(len(data) / epoch_len)
	return epoch_len, n_epochs

def get_epoch(data, i, epoch_len):
	"""
	Get the data corresponding to the `i`-th
	epoch.

	Parameters
	----------
	data : array-like, shape=(n_samples, n_features)
		Data to be epoched.

	i : int
		The epoch to collect.

	epoch_len : int
		Number of samples in each epoch.

	Returns
	-------
	data : ndarray, shape=(epoch_len,)
		The data of the `i`-th epoch.
	"""
	start = i * epoch_len
	end = (i + 1) * epoch_len
	return data[start:end]



"""Signal Processing Functions"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np
from scipy import signal

from sleepens.utils.data.normalize import prob_normalize

def compute_bands(f, Pxx, bands, compute='sum'):
	"""
	Compute the power spectrum within
	specified banding intervals.

	Parameters
	----------
	f : ndarray
		Array of frequencies.

	Pxx : ndarray
		Power spectral density of `data`.

	bands : dict
		Dictionary of band intervals to evaluate.

	compute : {'mean', 'median', 'max', 'min', 'sum'}, default='mean'
		Method to compute the summary statistic of the band.

	Returns
	-------
	dataBands : dict
		A dictionary of computed power values at each band,
		matching the bands provided in `bands`.
	"""
	data_bands = {}
	for band in bands.keys():
		f_ix = np.where((f >= bands[band][0]) & (f <= bands[band][1]))[0]
		if compute == 'mean':
			data_bands[band] = np.mean(Pxx[f_ix])
		elif compute == 'median':
			data_bands[band] = np.median(Pxx[f_ix])
		elif compute == 'max':
			data_bands[band] = np.max(Pxx[f_ix])
		elif compute == 'min':
			data_bands[band] = np.min(Pxx[f_ix])
		elif compute == 'sum':
			data_bands[band] = np.sum(Pxx[f_ix])
		else:
			raise ValueError("Provided compute argument must be 'mean','median','max','min', or 'sum'")
	return data_bands

def welch(data, fs=1.0, nperseg=None, noverlap=None, nfft=None, detrend='constant'):
	"""
	Estimate the power spectral density via Welch's method
	and Hann windows.

	Welch’s method computes an estimate of the power spectral
	density by dividing the data into overlapping segments,
	computing a modified periodogram for each segment and
	averaging the periodograms.

	Parameters
	----------
	data : array-like
		Time series of data values.

	fs : float, default=1.0
		Sampling frequency of `data`.

	nperseg : int or None, default=None
		Length of each segment.

	noverlap : int or None, default=None
		Number of samples to overlap between segments.
		If None, `noverlap` is half of `nperseg`.

	nfft : int or None, default=None
		Length of the FFT to be used. If None, it is
		set to `nperseg`.

	detrend : str or function or False, default='constant'
		Specifies detrend method for the data.
		If detrend is a string, it is passed as the type
		argument to the detrend function. If it is a function,
		it takes a segment and returns a detrended segment.
		If detrend is False, no detrending is done.

	Returns
	-------
	f : ndarray, shape=(n_frequencies,)
		Array of frequencies.

	Pxx : ndarray, shape=(n_frequencies,)
		Power spectral density of `data`.
	"""
	f, Pxx = signal.welch(data, fs=fs, nperseg=nperseg, noverlap=noverlap,
							nfft=nfft, detrend=detrend)
	return f, Pxx

def spectrogram(data, fs=1.0, nperseg=None, noverlap=None, nfft=None, detrend='constant'):
	"""
	Compute a spectrogram of consecutive Fourier Transforms,
	using a Turkey window of shape 0.25

	Parameters
	----------
	data : array-like
		Time series of data values.

	fs : float, default=1.0
		Sampling frequency of `data`.

	nperseg : int or None, default=None
		Length of each segment.

	noverlap : int or None, default=None
		Number of samples to overlap between segments.
		If None, `noverlap` is half of `nperseg`.

	nfft : int or None, default=None
		Length of the FFT to be used. If None, it is
		set to `nperseg`.

	detrend : str or function or False, default='constant'
		Specifies detrend method for the data.
		If detrend is a string, it is passed as the type
		argument to the detrend function. If it is a function,
		it takes a segment and returns a detrended segment.
		If detrend is False, no detrending is done.

	Returns
	-------
	f : ndarray, shape=(n_frequencies,)
		Array of frequencies.

	t : ndarray, shape=(n_segments,)
		Array of segment times.

	Sxx : ndarray, shape=(n_segments, n_frequencies)
		Spectrogram of `data`.
	"""
	f, t, Sxx = signal.spectrogram(data, fs=fs, nperseg=nperseg, noverlap=noverlap,
							nfft=nfft, detrend=detrend)
	return f, t, Sxx.T

def spectral_entropy(f, Pxx):
	"""
	Compute the spectral entropy of the power spectral density
	according to Shannon entropy.

	Parameters
	----------
	f : ndarray
		Array of frequencies.

	Pxx : ndarray
		Power spectral density.

	Returns
	-------
	entropy : float
		Spectral entropy between 0 and 1, with 0
		representing no entropy and 1 as maximal entropy.
	"""
	norm_Pxx = prob_normalize(Pxx)
	entropy = np.sum(norm_Pxx*np.log(1/norm_Pxx))
	normEntropy = entropy / np.log(len(f))
	return normEntropy

def rms(data):
	"""
	Calculate the root mean square of the data.

	Parameters
	----------
	data : array-like
		Time series of data values.

	Returns
	-------
	rms : float
		root mean square of the given data.
	"""
	return np.sqrt(np.mean(np.square(data)))

def zero_cross(data, zero=0):
	"""
	Calculate the number of times the data
	crosses the `zero` value.

	Parameters
	----------
	data : ndarray
		Time series of data values.

	zero : scalar, default=0
		Crossing point to analyze.

	Returns
	-------
	cross : float
		Number of times the data crosses `zero`.
	"""
	return (((data[:-1] - zero) * (data[1:] - zero)) < zero).sum()

def percentile_mean(data, k):
	"""
	Compute the mean of data above the k-th percentile.

	Parameters
	----------
	data : ndarray
		Time series of data values.

	k : float
		Percentile threshold.

	Returns
	-------
	mean : float
		Mean of the data above the k-th percentile.
	"""
	index = (len(data)*k)//100
	return np.mean(np.partition(data, index)[index:])



"""Normalization Functions"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.utils.data.transform import log_transform

def demean(data, axis=0):
	"""
	Demean the data by subtracting the mean from
	all samples.

	Parameters
	----------
	data : ndarray
		Data to be normalized.

	axis : int, default=0
		Axis to demean along.

	Returns
	-------
	demeaned : ndarray
		Demeaned data.
	"""
	return data - np.mean(data, axis=axis)

def normalize(data, mu=0, sigma=1, axis=0):
	"""
	Normalize `data` such that the mean is
	`mu` and standard deviation is `sigma`.

	Parameters
	----------
	data : ndarray
		Data to be normalized.

	mu : float, default=0
		Mean of the normalized data.

	sigma : float, default=1
		Standard deviation of the normalized data.

	axis : int, default=0
		Axis to normalize along.

	Returns
	-------
	norm : ndarray
		Normalized data.
	"""
	mean = np.mean(data, axis=axis) - mu
	std = np.std(data, axis=axis) * sigma
	return (data - mean) / std

def max_normalize(data, scale=(0,1), axis=0):
	"""
	Normalize `data` into an interval defined by `scale`.

	Parameters
	----------
	data : ndarray
		Data to be normalized.

	scale : tuple, shape=(2,), default=(0,1)
		Normalization interval. All data will be bounded
		within this interval such that the maximum of `data`
		is the maximum of `scale`.

	axis : int, default=0
		Axis to normalize along.

	Returns
	-------
	norm : ndarray
		Normalized data.
	"""
	factor = (scale[1] - scale[0]) / (data.max(axis=axis) - data.min(axis=axis))
	return factor * (data - data.min(axis=axis)) + scale[0]

def prob_normalize(data, scale=1, axis=0):
	"""
	Normalize `data` as a probability.
	Calculated as x divided by the sum of
	data values for each sample in the data.

	Parameters
	----------
	data : ndarray
		Data to be normalized.

	scale : scalar, default=1
		Normalization scale. A probability of 1
		corresponds to `scale`.

	axis : int, default=0
		Axis to normalize along.

	Returns
	-------
	norm : ndarray
		Normalized data.
	"""
	return (data / np.sum(data, axis=axis)) * scale

def log_normalize(data, base=np.e, scale=(0,1), axis=0):
	"""
	Log normalize `data`.

	Parameters
	----------
	data : ndarray
		Data to be normalized.

	base : {scalar, 'mean', 'median'}, default=np.e
		Log base for normalization. Must be one of:
		 - scalar : Constant base.
		 - 'mean' : Set the base to the mean of the data.
		 - 'median' : Set the base to the median of the data.

	scale : tuple, shape=(2,), default=(0,1)
		Normalization interval. All data will be bounded
		within this interval such that the maximum of `data`
		is the maximum of `scale`.

	axis : int, default=0
		Axis to normalize along.

	Returns
	-------
	norm : ndarray
		Normalized data.
	"""
	data = max_normalize(data, scale=(1,2), axis=axis)
	return max_normalize(log_transform(data, base=base, axis=axis),
							scale=scale, axis=axis)



"""Fast Fourier Transform Processor"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.io import DataObject, Dataset
from sleepens.utils.data import signal as s
from sleepens.utils import calculate_epochs, get_epoch

def fft(dataobject, epoch_size, name=None, nperseg_factor=1,
				noverlap_factor=0.1, detrend='constant'):
	"""
	Process the data in `dataobject`,
	divide the data into epochs of `epoch_size`, and
	compute the Fast Fourier Transform (FFT).

	Parameters
	----------
	dataobject : DataObject
		DataObject for processing.

	epoch_size : int
		Number of seconds in each epoch.

	nperseg_factor : float (0, inf)
		Multiplied by data resolution to
		calculate the NPERSEG value.

	noverlap_factor : float (0, inf)
		Multiplied by NPERSEG to calculate
		the NOVERLAP value.

	detrend : str or func or False, default='constant'
		Method to detrend each segment.

	Returns
	-------
	ds = Dataset
		Dataset with frequencies as features
		and power spectral densities at each epoch.
	"""
	f, Pxx = [], []
	epoch_len, n_epochs = calculate_epochs(dataobject.data, dataobject.resolution, epoch_size)
	for i in range(n_epochs):
		epoch = get_epoch(dataobject.data, i, epoch_len)
		fs = 1/dataobject.resolution
		nperseg = fs * nperseg_factor
		noverlap = nperseg * noverlap_factor
		fi, Pxxi = s.welch(epoch, fs=fs, nperseg=nperseg, noverlap=noverlap,
								detrend=detrend)
		f.append(fi)
		Pxx.append(Pxxi)
	return Dataset(name=name, features=f, data=Pxx)



"""MinREM Fix"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.postprocess import default_map

def REMSensitivity(Y_hat, p, map=default_map, avg_threshold=0.08, init_threshold=0.03,
				min_threshold=0.02, window=2, min_size=3):
	"""
	REM Sensitivity Addon. Increase the sensitivity of predictions
	to favour REM sleep episodes.

	Fix is conducted by moving in a backwards pass through the prediction.
	Upon encountering a REM probability at least the `init_threshold`,
	average the REM probabilities across the next `window` timepoints, inclusive.
	As long as the moving average meets the threshold and all REM probabilities
	are at least the `min_threshold` and the number of such timepoints is at least
	`min_size` in length, all such timepoints are overwritten as REM.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples, n_classes)
		The raw prediction probabilities.

	p : array-like, shape=(n_samples,)
		The predictions to process. If no Addon
		processing was done prior to this, `p`
		corresponds with Y_hat.

	map : dict
		Mapping the label values to some
		set of integers.

	avg_threshold : float, default=0.08
		The average across the window needed to trigger.

	init_threshold : float, default=0.03
		The minimum probability a state must have to begin
		triggering the averaging.

	min_threshold : float, default=0.02
		The minimum probability a state must have to allow
		averaging.

	window : int, default=2
		The window size for averaging.

	min_size : int, default=3
		The minimum size of a REM episode.

	Returns
	-------
	p : array-like, shape=(n_samples,)
		The post-processed predictions.
	"""
	rem_count = 0
	i = len(p) - window
	while i >= 0:
		if p[i] == map['R'] : rem_count += 1
		if rem_count == 0 and p[i+1] == map['NR'] and Y_hat[i,map['R']] >= init_threshold:
			avg = np.mean(Y_hat[i:i+window,map['R']])
			if avg >= avg_threshold : rem_count += 1
			else : rem_count = 0
		if rem_count > 0 and (p[i] != map['AW'] or p[i] != map['W']) and p[i] != map['R'] and Y_hat[i,map['R']] >= min_threshold:
			avg = np.mean(Y_hat[i:i+window,mao['R']])
			if avg >= avg_threshold:
				rem_count += 1
				if rem_count > min_size : p[i] = map['R']
			else : rem_count = 0
		elif p[i] != map['R'] : rem_count = 0
		if rem_count == min_size : p[i:i+min_size] = map['R']
		i -= 1
	return p



"""Calculate Functions"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

def calculate_batch(batch_size, length):
	"""
	Calculate the batch size for the data of given length.

	Parameters
	----------
	batch_size : int, float, default=None
		Batch size for training. Must be one of:
		 - int : Use `batch_size`.
		 - float : Use `batch_size * n_samples`.
		 - None : Use `n_samples`.

	length : int
		Length of the data to be batched.

	Returns
	-------
	batch : int
		Actual batch size.
	"""
	if batch_size is None : return length
	elif isinstance(batch_size, int) and batch_size > 0 and \
			batch_size <= length:
		return batch_size
	elif isinstance(batch_size, float) and 0 < batch_size <= 1:
		return int(batch_size * length)
	else:
		raise ValueError("Batch size must be None, an int less than %d," % length,
							"or a float within (0,1]")

def calculate_weight(Y, n_classes, class_weight=None, weights=None):
	"""
	Calculate the weights applied to the predicted labels,
	combining class weights and sample weights.

	Parameters
	----------
	Y : array-like, shape=(n_samples,)
		Target labels as integers.

	n_classes : int
		Number of classes.

	class_weight : dict, 'balanced', or None, default=None
		Weights associated with classes in the form
		`{class_label: weight}`. Must be one of:
		 - None : All classes have a weight of one.
		 - 'balanced': Class weights are automatically calculated as
						`n_samples / (n_samples * np.bincount(Y))`.

	weights : array-like, shape=(n_samples,), default=None
		Sample weights. If None, then samples are equally weighted.

	Returns
	-------
	weights : array-like, shape=(n_samples,)
		Weights combining sample weights and class weights.
	"""
	if class_weight is None and weights is None : return np.ones(len(Y))
	elif weights is None : weights = np.ones(len(Y))
	d = class_weight
	if isinstance(d, str) and d == 'balanced':
		l = len(Y) / (n_classes * np.bincount(Y))
		d = {k: l[k] for k in range(len(l))}
	if isinstance(d, dict):
		class_weights = np.array([d[k] for k in Y])
	elif d is None : class_weights = np.ones(len(Y))
	else : raise ValueError("Class Weight must either be a dict or 'balanced' or None")
	return weights * class_weights

def calculate_bootstrap(bootstrap_size, length):
	"""
	Calculate the bootstrap size for the data of given length.

	Parameters
	----------
	bootstrap_size : int, float, default=None
		Bootstrap size for training. Must be one of:
		 - int : Use `bootstrap_size`.
		 - float : Use `bootstrap_size * n_samples`.
		 - None : Use `n_samples`.

	length : int
		Length of the data to be bootstrapped.

	Returns
	-------
	bootstrap : int
		Actual bootstrap size.
	"""
	if bootstrap_size is None:
		return length
	elif isinstance(bootstrap_size, int) and bootstrap_size > 0:
		return bootstrap_size
	elif isinstance(bootstrap_size, float) and 0 < bootstrap_size <= 1:
		return int(bootstrap_size * length)
	else : raise ValueError("Bootstrap Size must be None, a positive int or float in (0,1]")



"""WakeSensitivity Fix"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.postprocess import default_map

def WakeSensitivity(Y_hat, p, map=default_map, threshold=0.5):
	"""
	Wake Sensitivity Addon. Increase the sensitivity
	of waking due to split of active and quiet waking periods.

	Fix is conducted by determining when the combined probability
	of the AW and QW states for a given timepoint sum to above the given
	threshold. When this condition is satisified, the prediction is overwritten
	to the waking state with the higher probability.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples, n_classes)
		The raw prediction probabilities.

	p : array-like, shape=(n_samples,)
		The predictions to process. If no Addon
		processing was done prior to this, `p`
		corresponds with Y_hat.

	map : dict
		Mapping the label values to some
		set of integers.

	threshold : float, [0,1), default=0.5
		The threshold for AW, QW states to reclassify
		as waking.

	Returns
	-------
	p : array-like, shape=(n_samples,)
		The post-processed predictions.
	"""
	for i in range(len(p)):
		if np.sum(Y_hat[i,:map['NR']]) >= threshold and p[i] >= map['NR']:
			p[i] = np.argmax(Y_hat[i,:map['NR']])
	return p



"""Classification Report"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.analysis import multiconfusion_matrix

def calculate_statistics(Y_hat, Y, beta=1, average=None):
	"""
	Calculate the precisions, recalls, F-beta scores, and
	supports for each class in `targets`.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples,)
		List of data labels.

	Y : array-like, shape=(n_samples,)
		List of target truth labels.

	beta : float, default=1
		Strength of recall relative to precision in the F-score.

	average : {'micro', 'macro', 'weighted', None}, default=None
		The type of averaging to perform on statistics. Must be one of:
		 - None : Do not perform averaging, statistics for each class
		 			are returned.
		 - 'micro' : Calculate globally, counting total true positives,
		 				false negatives, and false positives.
		 - 'macro' : Calculate per class an unweighted mean.
		 - 'weighted' : Calculate per class the mean weighted by support.

	Returns
	-------
	precisions : float or dict
		Dictionary of precisions for each class if `average` is None.
		Averaged precision based on averaging method if provided.

	recalls : float or dict
		Dictionary of recalls for each class if `average` is None.
		Averaged recall based on averaging method if provided.

	fscores : float or dict
		Dictionary of fscores for each class if `average` is None.
		Averaged fscore based on averaging method if provided.

	supports : float or dict
		Dictionary of supports for each class if `average` is None.
		Total support (number of classes) if averaging method is provided.
	"""
	if beta < 0:
		raise ValueError("Beta must be non-negative")
	matrix = multiconfusion_matrix(Y_hat, Y)
	matrix_labels = list(matrix.keys())
	matrix = np.array([matrix[l] for l in matrix_labels])
	tp_sum = matrix[:,1,1]
	label_sum = tp_sum + matrix[:,0,1]
	target_sum = tp_sum + matrix[:,1,0]
	if average == 'micro':
		tp_sum = np.array([tp_sum.sum()])
		label_sum = np.array([label_sum.sum()])
		target_sum = np.array([target_sum.sum()])
	with np.errstate(divide='ignore', invalid='ignore'):
		precisions = np.divide(tp_sum, label_sum,
								out=np.zeros(tp_sum.shape, dtype=float),
								where=label_sum!=0)
		recalls = np.divide(tp_sum, target_sum,
								out=np.zeros(tp_sum.shape, dtype=float),
								where=target_sum!=0)
	if np.isposinf(beta):
		fscores = recalls
	else:
		beta2 = beta ** 2
		denom = beta2 * precisions + recalls
		valid = np.where(denom != 0)[0]
		fscores = np.zeros_like(denom)
		fscores[valid] = (1 + beta2) * precisions[valid] * recalls[valid] / denom[valid]
	if average == 'weighted':
		weights = target_sum
		if target_sum.sum() == 0:
			return 0, 0, 0, target_sum.sum()
	else:
		weights = None
	if average is not None:
		precisions = np.average(precisions, weights=weights)
		recalls = np.average(recalls, weights=weights)
		fscores = np.average(fscores, weights=weights)
		supports = target_sum.sum()
	else:
		precisions = {matrix_labels[k]: precisions[k] for k in range(len(matrix_labels))}
		recalls = {matrix_labels[k]: recalls[k] for k in range(len(matrix_labels))}
		fscores = {matrix_labels[k]: fscores[k] for k in range(len(matrix_labels))}
		supports = {matrix_labels[k]: target_sum[k] for k in range(len(matrix_labels))}
	return precisions, recalls, fscores, supports

def classification_report(Y_hat, Y, beta=1):
	"""
	Create a report on classification statistics.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples,)
		List of data labels.

	Y : array-like, shape=(n_samples,)
		List of target truth labels.

	beta : float, default=1
		Strength of recall relative to precision in the F-score.

	Returns
	-------
	report : dict
		Dictionary containing classification statistics in the following
		structure:
		 - 'label': {
		 				'precision':0.5,
						'recall':1.0,
						'f-score':0.67,
						'support':1
		 			},
		   ...
		 - 'beta': 1,
		 - 'support': 5,
		 - 'accuracy': 0.8,
		 - 'macro avg': {
		 				'precision':0.6,
						'recall':0.9,
						'f-score':0.67,
		 			},
		 - 'weighted avg': {
		 				'precision':0.67,
						'recall':0.9,
						'f-score':0.67,
		 			}
	"""
	stats = calculate_statistics(Y_hat, Y, beta=beta)
	_, _, accuracy, total = calculate_statistics(Y_hat, Y, beta=beta, average='micro')
	macro = calculate_statistics(Y_hat, Y, beta=beta, average='macro')
	weighted = calculate_statistics(Y_hat, Y, beta=beta, average='weighted')
	h = ['precision', 'recall', 'f-score', 'support']
	report = {
				'beta': beta,
				'support': total,
				'accuracy': accuracy,
				'macro avg': {h[i]: macro[i] for i in range(len(h))},
				'weighted avg': {h[i]: weighted[i] for i in range(len(h))}
			}
	classes = set(stats[0].keys())
	for c in classes:
		report[c] = {h[i]: stats[i][c] for i in range(len(h))}
	return report



from sklearn.ensemble import GradientBoostingClassifier as gbm

from sleepens.ml._base_model import Classifier

class GradientBoostingClassifier(Classifier):
	"""
	Wrapper class for sci-kit learn's Gradient Boosting Classifier.

    Gradient Boosting Classifier builds an additive model in a
    forward stage-wise fashion; it allows for the optimization of
    arbitrary differentiable loss functions. In each stage `n_classes_`
    regression trees are fit on the negative gradient of the
    binomial or multinomial deviance loss function. Binary classification
    is a special case where only a single regression tree is induced.

    Parameters
    ----------
    loss : {'deviance', 'exponential'}, default='deviance'
        The loss function to be optimized. 'deviance' refers to
        deviance (= logistic regression) for classification
        with probabilistic outputs. For loss 'exponential' gradient
        boosting recovers the AdaBoost algorithm.

    learning_rate : float, default=0.1
        Learning rate shrinks the contribution of each tree by `learning_rate`.
        There is a trade-off between learning_rate and n_estimators.

    n_estimators : int, default=100
        The number of boosting stages to perform. Gradient boosting
        is fairly robust to over-fitting so a large number usually
        results in better performance.

    subsample : float, default=1.0
        The fraction of samples to be used for fitting the individual base
        learners. If smaller than 1.0 this results in Stochastic Gradient
        Boosting. `subsample` interacts with the parameter `n_estimators`.
        Choosing `subsample < 1.0` leads to a reduction of variance
        and an increase in bias.

    criterion : {'friedman_mse', 'mse'}, default='friedman_mse'
        The function to measure the quality of a split. Supported criteria
        are 'friedman_mse' for the mean squared error with improvement
        score by Friedman, 'mse' for mean squared error, and 'mae' for
        the mean absolute error. The default value of 'friedman_mse' is
        generally the best as it can provide a better approximation in
        some cases.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:
        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.
        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_depth : int, default=3
        The maximum depth of the individual regression estimators. The maximum
        depth limits the number of nodes in the tree. Tune this parameter
        for best performance; the best value depends on the interaction
        of the input variables.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.
        The weighted impurity decrease equation is the following::
            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)
        where `N` is the total number of samples, `N_t` is the number of
        samples at the current node, `N_t_L` is the number of samples in the
        left child, and `N_t_R` is the number of samples in the right child.
        `N`, `N_t`, `N_t_R` and `N_t_L` all refer to the weighted sum,
        if `sample_weight` is passed.

    init : estimator or 'zero', default=None
        An estimator object that is used to compute the initial predictions.
        `init` has to provide :meth:`fit` and :meth:`predict_proba`. If
        'zero', the initial raw predictions are set to zero. By default, a
        `DummyEstimator` predicting the classes priors is used.

    random_state : int, RandomState instance or None, default=None
        Controls the random seed given to each Tree estimator at each
        boosting iteration.
        In addition, it controls the random permutation of the features at
        each split. It also controls the random spliting of the training data
		to obtain a validation set if `n_iter_no_change` is not None.
        Pass an int for reproducible output across multiple function calls.

    max_features : {'auto', 'sqrt', 'log2'}, int or float, default=None
        The number of features to consider when looking for the best split:
        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `int(max_features * n_features)` features are considered at each
          split.
        - If 'auto', then `max_features=sqrt(n_features)`.
        - If 'sqrt', then `max_features=sqrt(n_features)`.
        - If 'log2', then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.
        Choosing `max_features < n_features` leads to a reduction of variance
        and an increase in bias.
        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than `max_features` features.

    verbose : int, default=0
        Enable verbose output. If 1 then it prints progress and performance
        once in a while (the more trees the lower the frequency). If greater
        than 1 then it prints progress and performance for every tree.

    max_leaf_nodes : int, default=None
        Grow trees with `max_leaf_nodes` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    warm_start : bool, default=False
        When set to `True`, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just erase the
        previous solution.

    validation_fraction : float, default=0.1
        The proportion of training data to set aside as validation set for
        early stopping. Must be between 0 and 1.
        Only used if `n_iter_no_change` is set to an integer.

    n_iter_no_change : int, default=None
        `n_iter_no_change` is used to decide if early stopping will be used
        to terminate training when validation score is not improving. By
        default it is set to None to disable early stopping. If set to a
        number, it will set aside `validation_fraction` size of the training
        data as validation and terminate training when validation score is not
        improving in all of the previous `n_iter_no_change` numbers of
        iterations. The split is stratified.

    tol : float, default=1e-4
        Tolerance for the early stopping. When the loss is not improving
        by at least tol for `n_iter_no_change` iterations (if set to a
        number), the training stops.

    ccp_alpha : non-negative float, default=0.0
    	Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        `ccp_alpha` will be chosen. By default, no pruning is performed.

    Attributes
    ----------
	gb : sklearn.ensemble.GradientBoostingClassifier
		The underlying GradientBoostingClassifier.

	n_classes_ : int
		Number of classes.

	n_features_ : int
		Number of features.

	Underlying Attributes
	---------------------
	n_estimators_ : int
    	The number of estimators as selected by early stopping (if
        `n_iter_no_change` is specified). Otherwise it is set to
        `n_estimators`.

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance.
        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values).

    oob_improvement_ : ndarray of shape (n_estimators,)
        The improvement in loss (= deviance) on the out-of-bag samples
        relative to the previous iteration.
        `oob_improvement_[0]` is the improvement in
        loss of the first stage over the `init` estimator.
        Only available if `subsample < 1.0`.

    train_score_ : ndarray of shape (n_estimators,)
        The i-th score `train_score_[i]` is the deviance (= loss) of the
        model at iteration `i` on the in-bag sample.
        If `subsample == 1` this is the deviance on the training data.

    loss_ : LossFunction
        The concrete `LossFunction` object.

    init_ : estimator
        The estimator that provides the initial predictions.
        Set via the `init` argument or `loss.init_estimator`.

    estimators_ : ndarray of DecisionTreeRegressor, shape=(n_estimators, `loss_.K`)
        The collection of fitted sub-estimators. `loss_.K` is 1 for binary
        classification, otherwise n_classes.

    classes_ : ndarray of shape (n_classes,)
        The classes labels.

    n_features_ : int
        The number of data features.

    n_classes_ : int
        The number of classes.

    max_features_ : int
        The inferred value of max_features.
	"""
	def __init__(self, loss='deviance', learning_rate=0.1, n_estimators=100,
				subsample=1.0, criterion='friedman_mse', min_samples_split=2,
				min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_depth=3,
				min_impurity_decrease=0.0, init=None,
				random_state=None, max_features=None, verbose=0, max_leaf_nodes=None,
				warm_start=False, validation_fraction=0.1, n_iter_no_change=None,
				tol=0.0001, ccp_alpha=0.0):
		Classifier.__init__(self, warm_start=warm_start, random_state=random_state, verbose=verbose)
		self.gb = gbm(loss=loss, learning_rate=learning_rate, n_estimators=n_estimators,
					subsample=subsample, criterion=criterion, min_samples_split=min_samples_split,
					min_samples_leaf=min_samples_leaf, min_weight_fraction_leaf=min_weight_fraction_leaf,
					max_depth=max_depth, min_impurity_decrease=min_impurity_decrease,
					init=init, random_state=random_state,
					max_features=max_features, verbose=verbose, max_leaf_nodes=max_leaf_nodes,
					warm_start=warm_start, validation_fraction=validation_fraction,
					n_iter_no_change=n_iter_no_change, tol=tol, ccp_alpha=ccp_alpha)

	def fit(self, X, Y):
		"""
		Train the classifier on the given data and labels.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Training data.

		Y : array-like, shape=(n_samples,)
			Target labels as integers.

		Returns
		-------
		self : Classifier
			Fitted classifier.
		"""
		X, Y = self._fit_setup(X, Y)
		return self.gb.fit(X, Y)

	def predict_proba(self, X):
		"""
		Return the prediction probabilities on the given data.

		Parameters
		----------
		X : array-like, shape=(n_samples, n_features)
			Data to predict.

		Returns
		-------
		Y_hat : array-like, shape=(n_samples, n_classes)
			Probability predictions.
		"""
		X = self._predict_setup(X)
		return self.gb.predict_proba(X)

	def _is_fitted(self):
		"""
		Returns if the Classifier has been trained and is
		ready to predict new data.

		Returns
		-------
		fitted : bool
			True if the Classifier is fitted, False otherwise.
		"""
		attributes = ["estimators_","n_classes_","n_features_"]
		return Classifier._is_fitted(self.gb, attributes=attributes)



"""Data Object"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

class DataObject:
	"""
	DataObject is a container to store some signal data.

	Parameters
	----------
	name : string, default=""
		Name of the DataObject.

	data : ndarray, default=()
		Signal data contained in the DataObject.

	resolution : float, default=-1
		Resolution of the signal data.

	length : int, default=-1
		Number of samples in the signal data.

	divide : int, default=-1
		Number of clock ticks per sample.

	Attributes
	----------
	indices : ndarray
		Array of indices corresponding to the signal data.
	"""
	def __init__(self, name="", data=np.zeros(0), resolution=-1, divide=1):
		self.name = name
		self.data = data
		self.indices = np.arange(len(data))
		self.resolution = resolution
		self.length = len(self.data)
		self.divide = divide

	def __str__(self):
		return type(self).__name__ + ": " + self.name

	def _get(self, arr, i, k=None):
		"""
		Return a subarray of `arr` at index `i`
		if `k` is None or from index `i`
		up to (non-exclusive) `k`.

		Parameters
		----------
		arr : array-like
			Array to be sliced.

		i : int
			Index of the first element of the returned slice.

		k : int or None, default=None
			End of the sliced subarray.

		Returns
		-------
		subarray : array-like
			Subarray of `arr`.
		"""
		if len(arr) == 0:
			return False
		if isinstance(k, int):
			return arr[i:k]
		else:
			return arr[i]

	def get_data(self, i, k=None):
		"""
		Get a subarray of the signal data from index
		`i` up to (non-exclusive) `k` or the end of the
		list.

		Parameters
		----------
		i : int
			Index of the first element of the returned slice.

		k : int or None, default=None
			End of the sliced subarray. End of the list if
			None.

		Returns
		-------
		subarray : array-like
			Subarray of data.
		"""
		return self._get(self.data, i, k)

	def get_indices(self, i, k=None):
		"""
		Get a subarray of indices from index
		`i` up to (non-exclusive) `k` or the end of the
		list.

		Parameters
		----------
		i : int
			Index of the first element of the returned slice.

		k : int or None, default=None
			End of the sliced subarray. End of the list if
			None.

		Returns
		-------
		subarray : array-like
			Subarray of indices.
		"""
		return self._get(self.indices, i, k)

	def process(self, function, arg=None):
		"""
		Process the signal data by a given function
		returning a new DataObject with processed data.

		Parameters
		----------
		function : function or None
			Defined function to process signal data.
			If None, return itself.

		arg : list or None, default=None
			Additional arguments for the defined function.

		Returns
		-------
		dataobject : DataObject
			DataObject with the signal data after process
			modification.
		"""
		if function is None:
			return self
		elif arg is None:
			data = function(self.data)
		else:
			data = function(self.data, arg)
		return DataObject(name=self.name, data=data, resolution=self.resolution)



from ._confusion_matrix import confusion_matrix, multiconfusion_matrix
from ._report import calculate_statistics, classification_report


__all__ = [
			'confusion_matrix',
			'multiconfusion_matrix',
			'calculate_statistics',
			'classification_report',
		]



"""SON I/O Interface"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np
from copy import deepcopy
from pathlib import Path
from sonpy import lib as sp

from sleepens.io import DataObject, Dataset

name = "SON"
standard = ".smr files exported by CED Spike2"
filetypes = [("SMR File", "*.smr")]
type = "RAW"
tags = {'r', 'w'}

def read_data(filepath, channel):
	"""
	Read the data file at a specific data channel.

	Parameters
	----------
	filepath : path
		Path to the .smr file.

	channel : str
		Name of the channel in the .smr file.

	Returns
	-------
	dataobject : DataObject
		The DataObject containing the data from
		the specific channel.
	"""
	sonfile = _load(filepath)
	channel_idx = _get_channel(sonfile, channel)
	max_time = sonfile.ChannelMaxTime(channel_idx)
	divide = sonfile.ChannelDivide(channel_idx)
	n = int(max_time / divide)
	data = sonfile.ReadFloats(channel_idx, n, 0, max_time)
	resolution = sonfile.GetTimeBase() * divide
	del sonfile
	return DataObject(name=channel, data=data, resolution=resolution, divide=divide)

def read_labels(filepath, channel, map={}):
	"""
	Read the data file at a specific label channel.

	Parameters
	----------
	filepath : path
		Path to the .mat file.

	channel : str
		Name of the channel in the .mat file.

	map : dict, default={}
		Mapping the label values to some
		set of integers. Not used with SON
		as codes are already in integers.

	Returns
	-------
	dataobject : DataObject
		The DataObject containing the labels from
		the specific channel.
	"""
	sonfile = _load(filepath)
	channel_idx = _get_channel(sonfile, channel)
	max_time = sonfile.ChannelMaxTime(channel_idx)
	divide = sonfile.ChannelDivide(channel_idx)
	resolution = sonfile.GetTimeBase() * divide
	labels = sonfile.ReadTextMarks(channel_idx, 2, 0, max_time)
	if len(labels) > 1:
		interval = labels[1].Tick - labels[0].Tick
		n = int(max_time / interval)
		labels = sonfile.ReadTextMarks(channel_idx, n, 0, max_time)
	for i in range(len(labels)):
		labels[i] = labels[i].Code1
	labels = np.array(labels).flatten()
	del sonfile
	return DataObject(name=channel, data=labels, resolution=resolution, divide=divide)

def write(filepath, dataobjects, labels, map={}, epoch_size=5):
	"""
	Write the dataset to a file.

	Parameters
	----------
	filepath : path
		Path to the .smr file to write.

	dataobjects : array-like of DataObject, shape=(n_channels,)
		DataObjects to write to the file.

	labels : DataObject
		Labels to write to the file.

	map : dict, default={}
		Mapping label integers to some set of label values.

	epoch_size : int, default=5
		Number of seconds in an epoch for labels.
	"""
	sonfile = sp.SonFile(filepath)
	if sonfile.GetOpenError() != 0:
		raise RunTimeError("Error writing .smr file:", sp.GetErrorString(sonfile.GetOpenError()))
	if not sonfile.CanWrite():
		raise RunTimeError("File is read only")
	num_channels = len(dataobjects)
	sonfile.SetTimeBase(labels.resolution)
	for i in range(num_channels):
		sonfile = _write_data(sonfile, dataobjects[i], i)
	tickbase = int(epoch_size / sonfile.GetTimeBase())
	if labels is not None:
		sonfile = _write_labels(sonfile, labels, num_channels, map, tickbase)
	del sonfile

def _write_data(sonfile, dataobject, channel_idx):
	"""
	Write an Adc channel into the SonFile.

	Parameters
	----------
	sonfile : SonFile
		SonFile object.

	dataobject : DataObject
		DataObject for writing.

	channel_idx : int
		The index of the channel to write to.

	Returns
	-------
	sonfile : SonFile
		SonFile file object.
	"""
	sonfile.SetWaveChannel(channel_idx, dataobject.divide, sp.DataType.RealWave)
	sonfile.SetChannelTitle(channel_idx, dataobject.name)
	sonfile.SetChannelUnits(channel_idx, 'volts')
	sonfile.SetChannelScale(channel_idx, 1.)
	sonfile.SetChannelOffset(channel_idx, 0.)
	sonfile.SetChannelYRange(channel_idx, -5., 5.)
	sonfile.WriteFloats(channel_idx, dataobject.data, 0)
	return sonfile

def _write_labels(sonfile, dataobject, channel_idx, map, tickbase):
	"""
	Write a TextMark channel into the SonFile.

	Parameters
	----------
	sonfile : SonFile
		SonFile object.

	dataobject : DataObject
		DataObject for writing.

	channel_idx : int
		The index of the channel to write to.

	map : dict, default={}
		Mapping label integers to some set of label values.

	epoch_size : int, default=5
		Number of seconds in an epoch for labels.

	Returns
	-------
	sonfile : SonFile
		SonFile file object.
	"""
	labels = deepcopy(dataobject.data)
	codes = deepcopy(dataobject.data)
	labels = [map[i] for i in labels]
	textmarks = np.empty(len(labels), dtype=sp.TextMarker)
	max_text_size = max(len(s) for s in labels) + 1
	sonfile.SetTextMarkChannel(channel_idx, 2, max_text_size)
	sonfile.SetChannelTitle(channel_idx, dataobject.name)
	tick = 0
	for i in range(len(labels)):
		textmarks[i] = sp.TextMarker(labels[i], nTick=tick, nCode1=codes[i])
		tick += tickbase
	sonfile.WriteTextMarks(channel_idx, textmarks)
	return sonfile

def _load(filepath):
	"""
	Attempt to load the .smr file.

	Parameters
	----------
	filepath : path
		Path to the .smr file.

	Returns
	-------
	sonfile : SonFile
		SonFile file object.
	"""
	sonfile = sp.SonFile(filepath, True)
	if sonfile.GetOpenError() != 0:
		raise FileNotFoundError("Error opening .smr file:", sp.GetErrorString(sonfile.GetOpenError()))
	return sonfile

def _get_channel(sonfile, channel):
	"""
	Find the specified channel index within the SonFile.

	Parameters
	----------
	sonfile : SonFile
		SonFile file object.

	channel : str
		Name of the channel in the .smr file.

	Returns
	-------
	channel_idx : int
		The index of the specified channel.
	"""
	max_channels = sonfile.MaxChannels()
	channels = []
	for i in range(max_channels):
		if sonfile.GetChannelTitle(i) == channel:
			return i
		elif sonfile.ChannelType(i) != sp.DataType.Off:
			channels.append(sonfile.GetChannelTitle(i))
	raise FileNotFoundError("Channel named " + channel + " not found. Instead found: " + str(channels))



"""Random State"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

def create_random_state(seed=None):
	"""
	Create a RandomState.

	Parameters
	----------
	seed : None or int or RandomState, default=None
		Initial seed for the RandomState. If `seed` is None,
		return the RandomState singleton. If `seed` is an int,
		return a RandomState with the seed set to the int.
		If `seed` is a RandomState, return that RandomState.

	Returns
	-------
	random_state : RandomState
		A RandomState object.
	"""
	if seed is None:
		return np.random.mtrand._rand
	elif isinstance(seed, (int, np.integer)):
		return np.random.RandomState(seed=seed)
	elif isinstance(seed, np.random.RandomState):
		return seed
	else:
		raise ValueError("Seed must be None, an int, or a Random State")



"""Confusion Matrices"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

def confusion_matrix(Y_hat, Y, norm=None):
	"""
	Calculate confusion matrix.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples,)
		List of data labels.

	Y : array-like, shape=(n_samples,)
		List of target truth labels.

	norm : {'label', 'target', 'all', None}, default=None
		Normalization on resulting matrix. Must be one of:
		 - 'label' : normalize on labels (columns).
		 - 'target' : normalize on targets (rows).
		 - 'all' : normalize on the entire matrix.
		 - None : No normalization.

	Returns
	-------
	matrix : ndarray, shape=(target_classes, label_classes)
		Confusion matrix with target classes as rows and
		label classes as columns. Classes are in sorted order.

	target_classes : list, shape=(target_classes,)
		List of the target classes (rows).

	label_classes : list, shape=(label_classes,)
		List of the label classes (rows).
	"""
	target_classes = sorted(set(Y))
	label_classes = sorted(set(Y_hat))
	target_dict = {target_classes[k]: k for k in range(len(target_classes))}
	label_dict = {label_classes[k]: k for k in range(len(label_classes))}
	matrix = np.zeros((len(target_classes), len(label_classes)))
	for label, target in zip(Y_hat, Y):
		matrix[target_dict[target],label_dict[label]] += 1
	if norm == 'label':
		matrix /= np.max(matrix, axis=0).reshape((1,matrix.shape[1]))
	elif norm == 'target':
		matrix /= np.max(matrix, axis=1).reshape((matrix.shape[0],1))
	elif norm == 'all':
		matrix /= np.max(matrix)
	elif norm is not None:
		raise ValueError("Norm must be one of {'label', 'target', 'all', None}")
	return matrix.astype(int), target_classes, label_classes

def multiconfusion_matrix(Y_hat, Y):
	"""
	Calculate confusion matrices for each class in `targets`.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples,)
		List of data labels.

	Y : array-like, shape=(n_samples,)
		List of target truth labels.

	Returns
	-------
	matrices : dict of ndarrays, shape=(2,2)
		Dictionary of 2x2 confusion matrices for each class,
		with true values as rows and label values as columns.
		Matrix is as follows:
			0	1
		0	TN	FP
		1	FN	TP
	"""
	target_classes = sorted(set(Y))
	target_dict = {target_classes[k]: k for k in range(len(target_classes))}
	matrices = np.zeros((len(target_classes),2,2))
	for label, target in zip(Y_hat, Y):
		if label == target:
			matrices[target_dict[target],1,1] += 1
			matrices[target_dict[target],0,0] -= 1
			matrices[:,0,0] += 1
		elif label != target and label in target_classes:
			matrices[target_dict[label],0,1] += 1
			matrices[target_dict[target],1,0] += 1
	matrices = {target_classes[k]: matrices[k] for k in range(len(target_classes))}
	return matrices



import numpy as np
from copy import deepcopy
from sklearn import metrics

from sleepens.protocols.sleepens4.params import params
from sleepens.protocols.sleepens4.processor import process as processor
from sleepens.ml import cross_validate
from sleepens.ml.models import StackedTimeSeriesEnsemble, GradientBoostingClassifier, TimeSeriesEnsemble
from sleepens import postprocess

class SleepEnsemble4:
	def __init__(self, params=params, verbose=0):
		self.name = "SleepEnsemble4"
		self.params = params
		self.verbose = verbose
		gbc = GradientBoostingClassifier(**params['classifier']['gbc'])
		tsens1 = TimeSeriesEnsemble(estimators=[deepcopy(gbc) for i in range(10)])
		tsens2 = TimeSeriesEnsemble(estimators=[deepcopy(gbc) for i in range(10)])
		self.classifier = StackedTimeSeriesEnsemble(layer_1=tsens1, layer_2=tsens2, warm_start=False,
								random_state=None, verbose=verbose)

	def read(self, reader, file, labels=False):
		eeg = reader.read_data(file, self.params['reader']['EEG_NAME'])
		neck = reader.read_data(file, self.params['reader']['NECK_NAME'])
		mass = reader.read_data(file, self.params['reader']['MASS_NAME'])
		if labels:
			labels = reader.read_labels(file, self.params['reader']['SCORE_NAME'],
									map=self.params['reader']['SCORE_MAP'])
		else : labels = None
		data = (eeg, neck, mass)
		return data, labels

	def process(self, data, labels, name):
		ds = processor(data, labels, name, self.params['process'], verbose=self.verbose)
		return ds

	def fit(self, X, Y):
		self.classifier.fit(X, Y)
		return self

	def predict(self, X):
		if not self.classifier._is_fitted():
			raise RuntimeError("Sleep Ensemble has not been fitted")
		Y_hat = self.classifier.predict_proba(X)
		if isinstance(Y_hat, list):
			p = [self._postprocess(y) for y in Y_hat]
		else : p = self._postprocess(Y_hat)
		return p, Y_hat

	def cross_validate(self, X, Y, cv=5, repeat=1, random_state=None,
						shuffle=True):
		if len(X) < cv:
			if len(X) == 1 : raise ValueError("X is length 1 and cross validation requires at least 2 samples/files")
			cv = len(X)
			if self.verbose > 1 : print("Found too small X, reducing cv to", cv)
		_, Y_hat = cross_validate(self.classifier, X, Y, cv=cv, repeat=repeat,
						random_state=random_state, shuffle=shuffle, verbose=self.verbose)
		if isinstance(Y_hat, list):
			p = [self._postprocess(y) for y in Y_hat]
		else:
			p = [self._postprocess(Y_hat)]
		return p, Y_hat

	def score(self, p, Y):
		if isinstance(p, list) and isinstance(Y, list):
			score = [metrics.accuracy_score(y_, p_) for p_, y_ in zip(p, Y)]
		elif not isinstance(p, list) and isinstance(Y, list):
			score = [metrics.accuracy_score(Y, p)]
		else : raise ValueError("Both p and Y must either be lists of ndarrays or ndarrays")
		return score

	def set_verbose(self, verbose):
		self.verbose = verbose
		self.classifier.set_verbose(verbose)

	def _postprocess(self, Y_hat):
		p = np.argmax(Y_hat, axis=1)
		p = postprocess.WakeMax(Y_hat, p, map=self.params['process']['INVERSE_MAP'])
		p = postprocess.WakeToREM(Y_hat, p, map=self.params['process']['INVERSE_MAP'])
		p = postprocess.REMDrop(Y_hat, p, map=self.params['process']['INVERSE_MAP'], window=3, threshold=0.1)
		p = postprocess.MinREM(Y_hat, p, map=self.params['process']['INVERSE_MAP'], min_rem=3)
		p = postprocess.TransitionFix(Y_hat, p, transitions=self.params['transition'])
		return p



"""Excel Worksheet I/O Interface"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np
import xlrd, xlwt
import os
from pathlib import Path

from sleepens.io import Dataset
from sleepens.utils import is_float, is_int

name = "XLS"
standard = ".xls dataset files"
filetypes = [("Microsoft Excel Worksheet", "*.xls")]
type = "DATASET"
tags = {'r', 'w'}

def read_data(filepath, cols, sheet=None, ):
	"""
	Read the data file at the specified columns.

	Parameters
	----------
	filepath : path
		Path to the .xls file.

	cols : array-like of int, shape=(n_features,)
		Indices of columns to read.

	sheet : int, str, None, default=None
		The sheet to use. If None, imports
		data in the first sheet only. If int, opens the
		i-th sheet. If string, opens the sheet by the given name.

	Returns
	-------
	dataset : Dataset
		The Dataset containing the data from
		the specified columns.
	"""
	ds = xlrd.open_workbook(filepath)
	if sheet is None : data_sheet = ds.sheet_by_name(ds.sheet_names()[0])
	elif isinstance(sheet, int) : data_sheet = ds.sheet_by_name(ds.sheet_names()[sheet])
	elif isinstance(sheet, str) : data_sheet = ds.sheet_by_name(sheet)
	header = np.array(data_sheet.row_values(0))
	features = header[cols]
	data = []
	rows = data_sheet.nrows
	for row in range(1, rows):
		data_row = []
		for col in cols:
			data_row.append(_read_cell(data_sheet, row, col))
		data.append(data_row)
	name = Path(filepath).stem
	return Dataset(name=name, features=features, data=data)

def read_labels(filepath, cols, sheet=None, map={}):
	"""
	Read the data file at the specified columns.

	Parameters
	----------
	filepath : path
		Path to the .xls file.

	cols : array-like of int, shape=(n_features,)
		Indices of columns to read.

	sheet : int, str, None, default=None
		The sheet to use. If None, imports
		data in the first sheet only. If int, opens the
		i-th sheet. If string, opens the sheet by the given name.

	map : dict, default={}
		Mapping the label values to some
		set of integers.

	Returns
	-------
	dataset : Dataset
		The Dataset containing the labels from
		the specified columns.
	"""
	ds = xlrd.open_workbook(filepath)
	if sheet is None : data_sheet = ds.sheet_by_name(ds.sheet_names()[0])
	elif isinstance(sheet, int) : data_sheet = ds.sheet_by_name(ds.sheet_names()[sheet])
	elif isinstance(sheet, str) : data_sheet = ds.sheet_by_name(sheet)
	header = np.array(data_sheet.row_values(0))
	label_names = header[cols]
	labels = []
	rows = data_sheet.nrows
	for row in range(1, rows):
		label_row= []
		for col in cols:
			cell = _read_cell(data_sheet, row, col)
			if map : label_row.append(map[cell])
			else : label_row.append(cell)
		labels.append(label_row)
	name = Path(filepath).stem
	return Dataset(name=name, label_names=label_names, labels=labels)

def write(filepath, dataset):
	"""
	Write the dataset to a file.

	Parameters
	----------
	filepath : path
		Path to the .xls file to write.

	dataset : Dataset
		Dataset to write.
	"""
	header = np.concatenate((dataset.features, dataset.label_names))
	header = header.reshape(1, len(header))
	if len(dataset.labels) > 0:
		labels = dataset.labels.reshape(-1,len(dataset.label_names))
		if len(dataset.data) > 0:
			contents = np.concatenate((dataset.data, labels), axis=1)
		else:
			contents = labels
	else:
		contents = dataset.data
	if header.shape[1] != contents.shape[1]:
		raise ValueError("Dimensions of header and data do not match")
	if os.path.isfile(filepath) : os.remove(filepath)
	wb = xlwt.Workbook()
	ws = wb.add_sheet("Results")
	output = np.concatenate((header, contents))
	for row in range(output.shape[0]):
		for col in range(output.shape[1]):
			ws.write(row, col, output[row][col])
	wb.save(filepath)

def _read_cell(data_sheet, row, col):
	"""
	Read a cell in a .xls file and convert to
	int or float where possible.

	Parameters
	----------
	data_sheet : Sheet
		The Sheet to read the data.

	row : int
		The row to read at.

	col : int
		The column to read at.

	Returns
	-------
	cell : string, int, float
		The data found at the given cell.
	"""
	try:
		cell = data_sheet.cell_value(row, col)
	except:
		print(row, col)
	if is_float(cell) : cell = float(cell)
	elif is_int(cell) : cell = int(cell)
	return cell
